import enum
from typing import Any, Dict, List, Optional, Union

from asynctnt.iproto import protocol

from .exceptions import TarantoolNotConnectedError
from .prepared import PreparedStatement
from .types import KeyType, MethodRet, SpaceType, TupleType


class _DbMock:
    def __getattr__(self, item):
        raise TarantoolNotConnectedError("Tarantool is not connected")


class Isolation(enum.IntEnum):
    DEFAULT = 0
    READ_COMMITTED = 1
    READ_CONFIRMED = 2
    BEST_EFFORT = 3


class Api:
    __slots__ = ("_db",)

    def __init__(self):
        self._db: Union[_DbMock, protocol.Db] = _DbMock()

    def _set_db(self, db: protocol.Db):
        self._db = db

    def _clear_db(self):
        self._db = _DbMock()

    def ping(self, *, timeout: float = -1.0) -> MethodRet:
        """
        Ping request coroutine

        :param timeout: Request timeout

        :returns: :class:`asynctnt.Response` instance
        """
        return self._db.ping(timeout=timeout)

    def call16(
        self,
        func_name: str,
        args: Optional[List[Any]] = None,
        *,
        timeout: float = -1.0,
        push_subscribe: bool = False,
    ) -> MethodRet:
        """
        Call16 request coroutine. It is a call with an old behaviour
        (return result of a Tarantool procedure is wrapped into a tuple,
        if needed)

        :param func_name: function name to call
        :param args: arguments to pass to the function (list object)
        :param timeout: Request timeout
        :param push_subscribe: Subscribe to push notifications

        :returns: :class:`asynctnt.Response` instance
        """
        return self._db.call16(
            func_name, args, timeout=timeout, push_subscribe=push_subscribe
        )

    def call(
        self,
        func_name: str,
        args: Optional[List[Any]] = None,
        *,
        timeout: float = -1.0,
        push_subscribe: bool = False,
    ) -> MethodRet:
        """
        Call request coroutine. It is a call with a new behaviour
        (return result of a Tarantool procedure is not wrapped into
        an extra tuple). If you're connecting to Tarantool with
        version < 1.7, then this call method acts like a call16 method

        Examples:

        .. code-block:: pycon

            # tarantool function:
            # function f(...)
            #     return ...
            # end

            >>> await conn.call('f')
            <Response sync=3 rowcount=0 data=[]>

            >>> await conn.call('f', [20, 42])
            <Response sync=3 rowcount=2 data=[20, 42]>

        :param func_name: function name to call
        :param args: arguments to pass to the function (list object)
        :param timeout: Request timeout
        :param push_subscribe: Subscribe to push notifications

        :returns: :class:`asynctnt.Response` instance
        """
        return self._db.call(
            func_name, args, timeout=timeout, push_subscribe=push_subscribe
        )

    def eval(
        self,
        expression: str,
        args: Optional[List[Any]] = None,
        *,
        timeout: float = -1.0,
        push_subscribe: bool = False,
    ) -> MethodRet:
        """
        Eval request coroutine.

        Examples:

        .. code-block:: pycon

            >>> await conn.eval('return 42')
            <Response sync=3 rowcount=1 data=[42]>


            >>> await conn.eval('return box.info.version')
            <Response sync=3 rowcount=1 data=['2.1.1-7-gd381a45b6']>

        :param expression: expression to execute
        :param args: arguments to pass to the function, that will
                     execute your expression (list object)
        :param timeout: Request timeout
        :param push_subscribe: Subscribe to push messages

        :returns: :class:`asynctnt.Response` instance
        """
        return self._db.eval(
            expression, args, timeout=timeout, push_subscribe=push_subscribe
        )

    def select(
        self,
        space: SpaceType,
        key: Optional[KeyType] = None,
        **kwargs,
    ) -> MethodRet:
        """
        Select request coroutine.

        Examples:

        .. code-block:: pycon

            >>> await conn.select('tester')
            <Response sync=3 rowcount=2 data=[
                <TarantoolTuple id=1 name='one'>,
                <TarantoolTuple id=2 name='two'>
            ]>

            >>> res = await conn.select('_space', ['tester'], index='name')
            >>> res.data
            [<TarantoolTuple id=512
                             owner=1
                             name='tester'
                             engine='memtx'
                             field_count=0
                             flags={}
                             format=[
                                {'name': 'id', 'type': 'unsigned'},
                                {'name': 'name', 'type': 'string'}
                             ]>]


        :param space: space id or space name.
        :param key: key to select
        :param offset: offset to use
        :param limit: limit to use
        :param index: index id or name
        :param iterator: one of the following

                    * iterator id (int number),
                    * :class:`asynctnt.Iterator` object
                    * string with an iterator name

        :param timeout: Request timeout

        :returns: :class:`asynctnt.Response` instance
        """
        return self._db.select(space, key, **kwargs)

    def insert(
        self,
        space: SpaceType,
        t: TupleType,
        *,
        replace: bool = False,
        timeout: float = -1,
    ) -> MethodRet:
        """
        Insert request coroutine.

        Examples:

        .. code-block:: pycon

            # Basic usage
            >>> await conn.insert('tester', [0, 'hello'])
            <Response sync=3 rowcount=1 data=[
                <TarantoolTuple id=0 name='hello'>
            ]>

            # Using dict as an argument tuple
            >>> await conn.insert('tester', {
            ...                     'id': 0
            ...                     'text': 'hell0'
            ...                   })
            <Response sync=3 rowcount=1 data=[
                <TarantoolTuple id=0 name='hello'>
            ]>

        :param space: space id or space name.
        :param t: tuple to insert (list object)
        :param replace: performs replace request instead of insert
        :param timeout: Request timeout

        :returns: :class:`asynctnt.Response` instance
        """
        return self._db.insert(space, t, replace=replace, timeout=timeout)

    def replace(
        self,
        space: SpaceType,
        t: TupleType,
        *,
        timeout: float = -1.0,
    ) -> MethodRet:
        """
        Replace request coroutine. Same as insert, but replace.

        :param space: space id or space name.
        :param t: tuple to insert (list object)
        :param timeout: Request timeout

        :returns: :class:`asynctnt.Response` instance
        """
        return self._db.replace(space, t, timeout=timeout)

    def delete(
        self,
        space: SpaceType,
        key: KeyType,
        **kwargs,
    ) -> MethodRet:
        """
        Delete request coroutine.

        Examples:

        .. code-block:: pycon

            # Assuming tuple [0, 'hello'] is in space tester

            >>> await conn.delete('tester', [0])
            <Response sync=3 rowcount=1 data=[
                <TarantoolTuple id=0 name='hello'>
            ]>

        :param space: space id or space name.
        :param key: key to delete
        :param index: index id or name
        :param timeout: Request timeout

        :returns: :class:`asynctnt.Response` instance
        """
        return self._db.delete(space, key, **kwargs)

    def update(
        self,
        space: SpaceType,
        key: KeyType,
        operations: List[Any],
        **kwargs,
    ) -> MethodRet:
        """
        Update request coroutine.

        Examples:

        .. code-block:: pycon

            # Assuming tuple [0, 'hello'] is in space tester

            >>> await conn.update('tester', [0], [ ['=', 1, 'hi!'] ])
            <Response sync=3 rowcount=1 data=[
                <TarantoolTuple id=0 name='hi!'>
            ]>

            # you can use fields names as well
            >>> res = await conn.update('tester', [0],
            ...                         [ ['=', 'text', 'hola'] ])
            <Response sync=3 rowcount=1 data=[
                <TarantoolTuple id=0 name='hola'>
            ]>

        :param space: space id or space name.
        :param key: key to update
        :param operations:
                Operations list of the following format:
                [ [op_type, field_no, ...], ... ]. Please refer to
                https://tarantool.org/doc/book/box/box_space.html?highlight=update#lua-function.space_object.update
                You can use field numbers as well as their names in space
                format as a field_no (if only fetch_schema is True).
                If field is unknown then TarantoolSchemaError is raised.
        :param index: index id or name
        :param timeout: Request timeout

        :returns: :class:`asynctnt.Response` instance
        """
        return self._db.update(space, key, operations, **kwargs)

    def upsert(
        self,
        space: SpaceType,
        t: TupleType,
        operations: List[Any],
        **kwargs,
    ) -> MethodRet:
        """
        Update request coroutine. Performs either insert or update
        (depending of either tuple exists or not)

        Examples:

        .. code-block:: pycon

            # upsert does not return anything
            >>> await conn.upsert('tester', [0, 'hello'],
            ...                   [ ['=', 1, 'hi!'] ])
            <Response sync=3 rowcount=0 data=[]>

        :param space: space id or space name.
        :param t: tuple to insert if it's not in space
        :param operations:
                Operations list to use for update if tuple is already in
                space. It has the same format as in update requets:
                [ [op_type, field_no, ...], ... ]. Please refer to
                https://tarantool.org/doc/book/box/box_space.html?highlight=update#lua-function.space_object.update
                You can use field numbers as well as their names in space
                format as a field_no (if only fetch_schema is True).
                If field is unknown then TarantoolSchemaError is raised.
        :param timeout: Request timeout

        :returns: :class:`asynctnt.Response` instance
        """
        return self._db.upsert(space, t, operations, **kwargs)

    def execute(
        self,
        query: Union[str, int],
        args: Optional[List[Union[Dict[str, Any], Any]]] = None,
        *,
        parse_metadata: bool = True,
        timeout: float = -1.0,
    ) -> MethodRet:
        """
        Executes an SQL statement (only for Tarantool > 2)

        Examples:

        .. code-block:: pycon

            >>> await conn.execute("select 1 as a, 2 as b")
            <Response sync=3 rowcount=1 data=[<TarantoolTuple A=1 B=2>]>

            >>> await conn.execute("select * from sql_space")
            <Response sync=3 rowcount=2 data=[
                <TarantoolTuple ID=1 NAME='James Bond'>,
                <TarantoolTuple ID=2 NAME='Ethan Hunt'>
            ]>

            >>> await conn.execute("select * from sql_space",
            ...                    parse_metadata=False)
            <Response sync=3 rowcount=2 data=[
                <TarantoolTuple 0=1 1='James Bond'>,
                <TarantoolTuple 0=2 1='Ethan Hunt'>
            ]>

        :param query: SQL query or statement_id
        :param args: Query arguments
        :param parse_metadata: Set to False to disable response's metadata
                               parsing for better performance
        :param timeout: Request timeout

        :returns: :class:`asynctnt.Response` instance
        """
        return self._db.execute(
            query, args, parse_metadata=parse_metadata, timeout=timeout
        )

    def prepare(self, query: str) -> PreparedStatement:
        """
        Create a :class:`asynctnt.prepared.PreparedStatement` instance
        :param query: query to be prepared
        """
        return PreparedStatement(self, query)

    def prepare_iproto(
        self,
        query: str,
        timeout: float = -1.0,
    ) -> MethodRet:
        """
        Low-level prepare() call
        :param query: query to be prepared
        :param timeout: request timeout
        """
        return self._db.prepare(query, timeout=timeout)

    def unprepare_iproto(
        self,
        stmt_id: int,
        timeout: float = -1.0,
    ) -> MethodRet:
        """
        Low-level unprepare() call
        :param stmt_id: query to be unprepared
        :param timeout: request timeout
        """
        return self._db.prepare(stmt_id, timeout=timeout)

    def begin(
        self,
        isolation: Isolation = Isolation.DEFAULT,
        tx_timeout: float = 0.0,
        timeout: float = -1.0,
    ) -> MethodRet:
        """
        Begin an interactive transaction within a stream
        :param isolation: isolation level
        :param tx_timeout: transaction timeout
        :param timeout: request timeout
        :return:
        """
        return self._db.begin(isolation.value, tx_timeout, timeout)

    def commit(self, timeout: float = -1.0) -> MethodRet:
        """
        Commit a running transaction
        :param timeout: request timeout
        """
        return self._db.commit(timeout)

    def rollback(self, timeout: float = -1.0) -> MethodRet:
        """
        Rollback a running transaction
        :param timeout: request timeout
        """
        return self._db.rollback(timeout)



from libc.stdint cimport uint16_t, uint32_t, uint64_t
from libc.string cimport memcpy


cdef inline uint64_t load_u64(const void * p):
    cdef:
        uint64_t res

    res = 0
    memcpy(&res, p, sizeof(res))
    return res

cdef inline uint64_t load_u32(const void * p):
    cdef:
        uint32_t res

    res = 0
    memcpy(&res, p, sizeof(res))
    return res

cdef inline uint64_t load_u16(const void * p):
    cdef:
        uint16_t res

    res = 0
    memcpy(&res, p, sizeof(res))
    return res

cdef inline void store_u64(void * p, uint64_t v):
    memcpy(p, &v, sizeof(v))

cdef inline void store_u32(void * p, uint32_t v):
    memcpy(p, &v, sizeof(v))

cdef inline void store_u16(void * p, uint16_t v):
    memcpy(p, &v, sizeof(v))



cimport cython


@cython.final
cdef class EvalRequest(BaseRequest):
    cdef int encode_body(self, WriteBuffer buffer) except -1:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len

            bytes expression_temp
            char *expression_str
            ssize_t expression_len

        expression_str = NULL
        expression_len = 0

        expression_temp = encode_unicode_string(self.expression, buffer._encoding)
        cpython.bytes.PyBytes_AsStringAndSize(expression_temp,
                                              &expression_str,
                                              &expression_len)
        body_map_sz = 2
        # Size description:
        # mp_sizeof_map()
        # + mp_sizeof_uint(TP_EXPRESSION)
        # + mp_sizeof_str(expression)
        # + mp_sizeof_uint(TP_TUPLE)
        max_body_len = 1 \
                       + 1 \
                       + mp_sizeof_str(<uint32_t> expression_len) \
                       + 1

        buffer.ensure_allocated(max_body_len)

        p = begin = &buffer._buf[buffer._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_EXPR)
        p = mp_encode_str(p, expression_str, <uint32_t> expression_len)

        p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
        buffer._length += (p - begin)
        p = encode_key_sequence(buffer, p, self.args, None, False)



import logging

from asynctnt import Iterator, Response
from asynctnt.exceptions import TarantoolSchemaError
from tests import BaseTarantoolTestCase
from tests.util import get_complex_param


class SelectTestCase(BaseTarantoolTestCase):
    LOGGING_LEVEL = logging.INFO

    async def _fill_data(self, count=3, space=None):
        space = space or self.TESTER_SPACE_ID
        data = []
        for i in range(count):
            t = [i, str(i), 1, 2, "something"]
            data.append(t)
            await self.conn.insert(space, t)
        return data

    async def _fill_data_dict(self, count=3):
        data = []
        for i in range(count):
            t = {
                "f1": i,
                "f2": str(i),
                "f3": 1,
                "f4": 2,
                "f5": "something",
            }
            t = await self.conn.insert(self.TESTER_SPACE_ID, t)
            data.append(dict(t[0]))
        return data

    async def test__select_by_id_empty_space(self):
        res = await self.conn.select(self.TESTER_SPACE_ID)

        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [], "Body ok")

    async def test__select_by_id_non_empty_space(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_ID)

        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, data, "Body ok")

    async def test__select_by_name_space_empty(self):
        res = await self.conn.select(self.TESTER_SPACE_NAME)

        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [], "Body ok")

    async def test__select_by_name_non_empty_space(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_NAME)

        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, data, "Body ok")

    async def test__select_by_index_id(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_ID, index=1)

        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, data, "Body ok")

    async def test__select_by_index_name(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_ID, index="txt")

        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, data, "Body ok")

    async def test__select_by_id_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        try:
            await self.conn.select(self.TESTER_SPACE_ID)
        except Exception as e:
            self.fail(e)

    async def test__select_by_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.select(self.TESTER_SPACE_NAME)

    async def test__select_by_index_id_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        try:
            await self.conn.select(self.TESTER_SPACE_ID, index=1)
        except Exception as e:
            self.fail(e)

    async def test__select_by_index_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.select(self.TESTER_SPACE_NAME, index="txt")

    async def test__select_by_key_one_item(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_NAME, [1])
        self.assertResponseEqual(res, [data[1]], "Body ok")

    async def test__select_by_key_multiple_items_index(self):
        data = await self._fill_data()
        next_id = data[-1][0] + 1
        next_txt = data[-1][1]
        await self.conn.insert(self.TESTER_SPACE_ID, [next_id, next_txt, 1, 2, "text"])
        data.append([next_id, next_txt, 1, 2, "text"])

        res = await self.conn.select(self.TESTER_SPACE_NAME, [next_txt], index="txt")
        self.assertResponseEqual(res, data[len(data) - 2 :], "Body ok")

    async def test__select_limit(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_NAME, limit=1)
        self.assertResponseEqual(res, [data[0]], "Body ok")

    async def test__select_limit_offset(self):
        data = await self._fill_data(4)

        res = await self.conn.select(self.TESTER_SPACE_NAME, limit=1, offset=2)
        self.assertResponseEqual(res, [data[2]], "Body ok")

    async def test__select_iterator_class(self):
        data = await self._fill_data(4)

        res = await self.conn.select(self.TESTER_SPACE_NAME, iterator=Iterator.GE)
        self.assertResponseEqual(res, data, "Body ok")

        res = await self.conn.select(self.TESTER_SPACE_NAME, iterator=Iterator.LE)
        self.assertResponseEqual(res, list(reversed(data)), "Body ok")

    async def test__select_iterator_int(self):
        data = await self._fill_data(4)

        res = await self.conn.select(self.TESTER_SPACE_NAME, iterator=4)
        self.assertResponseEqual(res, list(reversed(data)), "Body ok")

    async def test__select_iterator_str(self):
        data = await self._fill_data(4)

        res = await self.conn.select(self.TESTER_SPACE_NAME, iterator="LE")
        self.assertResponseEqual(res, list(reversed(data)), "Body ok")

    async def test__select_complex(self):
        p, p_cmp = get_complex_param(replace_bin=False)
        data = [1, "hello2", 1, 4, p_cmp]

        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.select(self.TESTER_SPACE_ID)
        self.assertResponseEqual(res, [data], "Body ok")

    async def test__select_all_params(self):
        data = await self._fill_data(10)

        res = await self.conn.select(
            self.TESTER_SPACE_NAME,
            index="primary",
            limit=2,
            offset=1,
            iterator=Iterator.LE,
        )
        self.assertResponseEqual(res, list(reversed(data))[1:3], "Body ok")

    async def test__select_all_by_hash_index(self):
        data = await self._fill_data(4, space="no_schema_space")

        res = await self.conn.select("no_schema_space", index="primary_hash")
        self.assertResponseEqual(res, data, "Body ok")

    async def test__select_key_tuple(self):
        try:
            await self.conn.select(self.TESTER_SPACE_ID, (1,))
        except Exception as e:
            self.fail(e)

    async def test__select_invalid_types(self):
        with self.assertRaisesRegex(
            TypeError, r"missing 1 required positional argument: \'space\'"
        ):
            await self.conn.select()

        with self.assertRaisesRegex(
            TypeError, r"sequence must be either list, tuple or dict"
        ):
            await self.conn.select(self.TESTER_SPACE_ID, 1)

        with self.assertRaisesRegex(TypeError, r"Index must be either str or int, got"):
            await self.conn.select(self.TESTER_SPACE_ID, [1], index=[1, 2])

        with self.assertRaisesRegex(TypeError, r"an integer is required"):
            await self.conn.select(self.TESTER_SPACE_ID, [1], index=1, limit="hello")

        with self.assertRaisesRegex(TypeError, r"an integer is required"):
            await self.conn.select(
                self.TESTER_SPACE_ID, [1], index=1, limit=1, offset="hello"
            )

        with self.assertRaisesRegex(TypeError, r"Iterator is of unsupported type"):
            await self.conn.select(
                self.TESTER_SPACE_ID, [1], index=1, limit=1, offset=1, iterator=[1, 2]
            )

    async def test__select_dict_key(self):
        data = await self._fill_data()
        res = await self.conn.select(self.TESTER_SPACE_ID, {"f1": data[0][0]})
        self.assertResponseEqual(res, [data[0]], "Body ok")

    async def test__select_dict_key_wrong_field(self):
        data = await self._fill_data()
        res = await self.conn.select(self.TESTER_SPACE_ID, {"f2": data[0][0]})
        self.assertResponseEqual(res, data, "Body ok")

    async def test__select_dict_key_other_index(self):
        data = await self._fill_data()
        res = await self.conn.select(
            self.TESTER_SPACE_ID, {"f2": data[0][1]}, index="txt"
        )
        self.assertResponseEqual(res, [data[0]], "Body ok")

    async def test__select_dict_resp(self):
        data = await self._fill_data_dict()
        res = await self.conn.select(self.TESTER_SPACE_ID, [])
        self.assertResponseEqualKV(res, data)



import warnings

from asynctnt import TarantoolTuple
from tests import BaseTarantoolTestCase


class ResponseTestCase(BaseTarantoolTestCase):
    async def _fill_data(self, count=3):
        data = []
        for i in range(count):
            t = [i, str(i), 1, 2, "something"]
            data.append(t)
            await self.conn.insert(self.TESTER_SPACE_ID, t)
        return data

    async def _fill_data_dict(self, count=3):
        data = []
        for i in range(count):
            t = {
                "f1": i,
                "f2": str(i),
                "f3": 1,
                "f4": 2,
                "f5": "something",
            }
            t = await self.conn.insert(self.TESTER_SPACE_ID, t)
            data.append(dict(t[0]))
        return data

    async def test__response_indexing(self):
        res = await self.conn.call("box.info")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            self.assertEqual(len(res), len(res.body), "len ok")
            self.assertEqual(res[0], res.body[0], "value ok")

    async def test__response_iter(self):
        data = await self._fill_data(3)
        res = await self.conn.select(self.TESTER_SPACE_ID)

        self.assertEqual(len(res), len(data), "len ok")

        res_arr = []
        for el in res:
            res_arr.append(list(el))
        self.assertListEqual(res_arr, data, "list ok")

    async def test__response_tuple_iter(self):
        data = await self._fill_data(1)
        res = await self.conn.select(self.TESTER_SPACE_ID)
        t = res[0]

        t_list = [el for el in t]  # check iteration over tuple  # noqa: C416
        self.assertListEqual(t_list, data[0], "tuple ok")

    async def test__response_tuple_keys(self):
        await self._fill_data(1)
        res = await self.conn.select(self.TESTER_SPACE_ID)
        t = res[0]

        correct_keys = ["f1", "f2", "f3", "f4", "f5"]
        self.assertListEqual(list(t.keys()), correct_keys, "keys ok")

    async def test__response_tuple_values(self):
        data = await self._fill_data(1)
        res = await self.conn.select(self.TESTER_SPACE_ID)
        t = res[0]

        self.assertListEqual(list(t.values()), data[0], "values ok")

    async def test__response_tuple_items(self):
        data = [0, "hello", 5, 6, "help", "common", "yo"]
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        t = res[0]
        d = {
            "f1": data[0],
            "f2": data[1],
            "f3": data[2],
            "f4": data[3],
            "f5": data[4],
        }

        t_dict = {k: v for k, v in t.items()}  # noqa: C416
        self.assertDictEqual(t_dict, d, "items ok")

    async def test__response_tuple_dict_extra_index(self):
        data = [0, "hello", 5, 6, "help", "common", "yo"]
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertEqual(res[0], data[0])
        self.assertEqual(res[1], data[1])
        self.assertEqual(res[2], data[2])
        self.assertEqual(res[3], data[3])
        self.assertEqual(res[4], data[4])
        self.assertEqual(res[5], data[5])
        self.assertEqual(res[6], data[6])
        self.assertEqual(res[-1], data[-1])
        self.assertEqual(res[-3], data[-3])

        self.assertEqual(res["f1"], data[0])
        self.assertEqual(res["f2"], data[1])
        self.assertEqual(res["f3"], data[2])
        self.assertEqual(res["f4"], data[3])
        self.assertEqual(res["f5"], data[4])

    async def test__response_tuple_slice(self):
        data = [0, "hello", 5, 6, "help", "common", "yo"]
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertEqual(type(res[:3]), tuple)

        self.assertListEqual(list(res[:3]), data[:3])
        self.assertListEqual(list(res[1:5]), data[1:5])
        self.assertListEqual(list(res[5:20]), data[5:20])
        self.assertEqual(list(res[7:3:-1]), data[7:3:-1])
        self.assertEqual(list(res[7:3:-2]), data[7:3:-2])

    async def test__response_tuple_contains(self):
        data = [0, "hello", 5, 6, "help", "common", "yo"]
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertTrue("f1" in res)
        self.assertTrue("f2" in res)
        self.assertTrue("f3" in res)
        self.assertTrue("f4" in res)
        self.assertTrue("f5" in res)
        self.assertFalse("f6" in res)

    async def test__response_tuple_key_error(self):
        data = [0, "hello", 5, 6, "help", "common", "yo"]
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        with self.assertRaises(KeyError):
            # noinspection PyStatementEffect
            res["f100"]

    async def test__response_tuple_get(self):
        data = [0, "hello", 5, 6, "help", "common", "yo"]
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertEqual(res.get("f1"), 0)
        self.assertEqual(res.get("f2"), "hello")
        self.assertEqual(res.get("f100"), None)
        self.assertEqual(res.get("f100", "zz"), "zz")

    async def test__response_with_no_space_format(self):
        res = await self.conn.insert("no_schema_space", [0, "one"])

        with self.assertRaises(ValueError):
            res[0].keys()

        with self.assertRaises(ValueError):
            res[0].items()

    async def test__native_response_with_no_space_format(self):
        await self.conn.insert("no_schema_space", [0, "one"])

        res = await self.conn.select("no_schema_space")
        self.assertEqual(1, res.rowcount, "count correct")
        self.assertTrue(
            isinstance(res[0], TarantoolTuple), "expecting a TarantoolTuple"
        )
        self.assertResponseEqual(res, [[0, "one"]], "resp ok")

        try:
            repr(res)
        except Exception as e:
            self.fail(e)

    async def test__response_repr(self):
        data = [0, "hello", 5, 6, "help", "common", "yo"]
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.select("tester")
        self.assertEqual(1, res.rowcount, "count correct")
        self.assertTrue(
            isinstance(res[0], TarantoolTuple), "expecting a TarantoolTuple"
        )

        self.assertEqual(
            "<TarantoolTuple f1=0 f2='hello' f3=5 f4=6 f5='help' " "5='common' 6='yo'>",
            repr(res[0]),
            "repr ok",
        )

    async def test__response_repr_trunc(self):
        data = [0, "hello", 5, 6, "help", "common", "yo"]
        for _ in range(50):
            data.append("x")

        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.select("tester")
        self.assertEqual(1, res.rowcount, "count correct")
        self.assertTrue(
            isinstance(res[0], TarantoolTuple), "expecting a TarantoolTuple"
        )

        tail = []
        for i in range(7, 50):  # 50: maximum number of fields to show in repr
            tail.append(f"{i}={repr('x')}")

        self.assertEqual(
            f"<TarantoolTuple f1=0 f2='hello' f3=5 f4=6 f5='help' 5='common' 6='yo' {' '.join(tail)} ...>",
            repr(res[0]),
            "repr ok",
        )

    async def test__response_str(self):
        data = [0, "hello", 5, 6, "help", "common", "yo"]
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.select("tester")
        self.assertEqual(1, res.rowcount, "count correct")
        self.assertTrue(
            isinstance(res[0], TarantoolTuple), "expecting a TarantoolTuple"
        )

        try:
            str(res)
        except Exception as e:
            self.fail(e)

    async def test__metadata(self):
        self.assertIsNotNone(self.conn.schema.id)
        self.assertIsNotNone(self.conn.schema.spaces)
        self.assertIn(self.TESTER_SPACE_NAME, self.conn.schema.spaces)
        self.assertIn(self.TESTER_SPACE_ID, self.conn.schema.spaces)
        self.assertIs(
            self.conn.schema.spaces[self.TESTER_SPACE_NAME],
            self.conn.schema.spaces[self.TESTER_SPACE_ID],
        )

        sp = self.conn.schema.spaces[self.TESTER_SPACE_NAME]
        self.assertEqual(self.TESTER_SPACE_NAME, sp.name)
        self.assertEqual(self.TESTER_SPACE_ID, sp.sid)
        self.assertEqual("memtx", sp.engine)
        self.assertEqual(4, len(sp.indexes))
        self.assertIn("primary", sp.indexes)
        self.assertIn(0, sp.indexes)
        self.assertIn("txt", sp.indexes)
        self.assertIn(1, sp.indexes)
        self.assertEqual(5, len(sp.metadata.fields))
        self.assertEqual("f1", sp.metadata.fields[0].name)
        self.assertEqual("unsigned", sp.metadata.fields[0].type)
        self.assertEqual("f2", sp.metadata.fields[1].name)
        self.assertEqual("string", sp.metadata.fields[1].type)
        self.assertEqual("f5", sp.metadata.fields[4].name)
        self.assertEqual("*", sp.metadata.fields[4].type)
        self.assertEqual(5, len(sp.metadata.name_id_map))
        self.assertEqual(0, sp.metadata.name_id_map["f1"])

        idx = sp.indexes[0]
        self.assertEqual(0, idx.iid)
        self.assertEqual("primary", idx.name)
        self.assertEqual(self.TESTER_SPACE_ID, idx.sid)
        self.assertEqual("tree", idx.index_type)
        self.assertEqual(1, len(idx.metadata.fields))
        self.assertEqual("f1", idx.metadata.fields[0].name)
        self.assertEqual("unsigned", idx.metadata.fields[0].type)
        self.assertEqual(1, len(idx.metadata.name_id_map))
        self.assertEqual(0, idx.metadata.name_id_map["f1"])

    async def test__metadata_is_nullable(self):
        sp_name = "test_space_with_nullable"
        await self.conn.eval(
            """
            local s = box.schema.space.create('%s')
            s:format({
                {name = "id", type = "unsigned"},
                {name = "name", type = "string", is_nullable = true},
            })
        """
            % (sp_name,)
        )

        try:
            # just to be sure that schema is refreshed
            await self.conn.refetch_schema()

            self.assertIn(sp_name, self.conn.schema.spaces)
            sp = self.conn.schema.spaces[sp_name]
            self.assertEqual(sp_name, sp.name)
            self.assertEqual("memtx", sp.engine)
            self.assertEqual(0, len(sp.indexes))
            self.assertEqual(2, len(sp.metadata.fields))
            self.assertEqual("id", sp.metadata.fields[0].name)
            self.assertEqual("unsigned", sp.metadata.fields[0].type)
            self.assertEqual(None, sp.metadata.fields[0].is_nullable)
            self.assertEqual("name", sp.metadata.fields[1].name)
            self.assertEqual("string", sp.metadata.fields[1].type)
            self.assertEqual(True, sp.metadata.fields[1].is_nullable)
            self.assertEqual(2, len(sp.metadata.name_id_map))
            self.assertEqual(0, sp.metadata.name_id_map["id"])
            self.assertEqual(1, sp.metadata.name_id_map["name"])

        finally:
            await self.conn.eval(
                """
                local s = box.space['%s']
                if s ~= nil then
                    s:drop()
                end
            """
                % (sp_name,)
            )



# cython: profile=True

from asynctnt.exceptions import TarantoolSchemaError
from asynctnt.log import logger

cimport cpython.list
cimport cython
from cpython.ref cimport PyObject
from libc.stdint cimport int32_t, uint64_t


@cython.final
cdef class Field:
    def __cinit__(self):
        self.name = None
        self.type = None
        self.collation = None
        self.is_nullable = None
        self.is_autoincrement = None
        self.span = None

    def __repr__(self):  # pragma: nocover
        return "<Field name={} type={} is_nullable={}>".format(
            self.name, self.type, self.is_nullable
        )

@cython.final
@cython.freelist(METADATA_FREELIST_SIZE)
cdef class Metadata:
    def __cinit__(self):
        self.fields = []
        self.names = []
        self.name_id_map = {}

    cdef inline void add(self, int id, Field field):
        cpython.list.PyList_Append(self.fields, field)
        cpython.list.PyList_Append(self.names, field.name)
        cpython.dict.PyDict_SetItem(self.name_id_map, field.name, id)

    cdef inline str name_by_id(self, int id):
        cdef Field field
        field = <Field> cpython.list.PyList_GetItem(self.fields, id)
        return field.name

    cdef inline int id_by_name(self, str name) except *:
        cdef:
            PyObject *fld

        fld = cpython.dict.PyDict_GetItem(self.name_id_map, name)
        if fld == NULL:
            raise KeyError('Field \'{}\' not found'.format(name))
        return <int> <object> fld

    cdef inline int id_by_name_safe(self, str name) except *:
        cdef:
            PyObject *fld

        fld = cpython.dict.PyDict_GetItem(self.name_id_map, name)
        if fld == NULL:
            return -1
        return <int> <object> fld

    cdef inline int len(self):
        return <int> cpython.list.PyList_GET_SIZE(self.fields)

    def __repr__(self):  # pragma: nocover
        return '<Metadata [fields_count={}]>'.format(self.len())

@cython.final
cdef class SchemaIndex:
    def __cinit__(self):
        self.iid = -1
        self.sid = -1
        self.name = None
        self.index_type = None
        self.unique = None
        self.metadata = None

    def __repr__(self):  # pragma: nocover
        return \
            '<{} sid={}, id={}, name={}, ' \
            'type={}, unique={}>'.format(
                self.__class__.__name__,
                self.sid, self.iid, self.name, self.index_type, self.unique
            )

@cython.final
cdef class SchemaSpace:
    def __cinit__(self):
        self.sid = -1
        self.owner = -1
        self.name = None
        self.engine = None
        self.field_count = 0
        self.flags = None

        self.metadata = None
        self.indexes = {}

    cdef void add_index(self, SchemaIndex idx):
        cpython.dict.PyDict_SetItem(self.indexes, idx.iid, idx)
        if idx.name:
            cpython.dict.PyDict_SetItem(self.indexes, idx.name, idx)

    cdef SchemaIndex get_index(self, index, create_dummy=True):
        cdef:
            SchemaIndex idx
            bint is_str, is_int
            PyObject *obj_p
        is_str = isinstance(index, str)
        is_int = isinstance(index, int)
        if not is_str and not is_int:
            raise TypeError(
                'Index must be either str or int, got: {}'.format(type(index)))

        obj_p = cpython.dict.PyDict_GetItem(self.indexes, index)
        if obj_p is not NULL:
            return <SchemaIndex> obj_p
        else:
            if is_int and create_dummy:
                logger.debug(
                    'Index %s not found in space %s/%s. Creating dummy.',
                    index, self.sid, self.name
                )
                idx = <SchemaIndex> SchemaIndex.__new__(SchemaIndex)
                idx.iid = index
                idx.sid = self.sid
                idx.name = str(index)
                cpython.dict.PyDict_SetItem(self.indexes, index, idx)
                return idx
            else:
                raise TarantoolSchemaError(
                    'Index {} not found in space {}/{}'.format(
                        index, self.sid, self.name
                    )
                )

    def __repr__(self):  # pragma: nocover
        return '<{} id={} name={} engine={}>'.format(
            self.__class__.__name__,
            self.sid, self.name, self.engine
        )

@cython.final
cdef class Schema:
    def __cinit__(self, int schema_id):
        self.id = schema_id
        self.spaces = {}

    cdef SchemaSpace get_space(self, space):
        cdef PyObject *obj_p = \
            cpython.dict.PyDict_GetItem(self.spaces, space)
        if obj_p is NULL:
            return None
        return <SchemaSpace> obj_p

    cdef SchemaSpace get_or_create_space(self, space):
        cdef:
            bint is_str, is_int
            PyObject *obj_p
        is_str = isinstance(space, str)
        is_int = isinstance(space, int)
        if not is_str and not is_int:
            raise TypeError(
                'Space must be either str or int, got: {}'.format(type(space)))

        obj_p = cpython.dict.PyDict_GetItem(self.spaces, space)
        if obj_p is NULL:
            if is_str:
                raise TarantoolSchemaError(
                    'Space {} not found'.format(space)
                )
            else:
                return self.create_dummy_space(space)
        return <SchemaSpace> obj_p

    cdef SchemaSpace create_dummy_space(self, int space_id):
        cdef SchemaSpace s
        logger.debug('Space %s not found. Creating dummy.', space_id)
        s = <SchemaSpace> SchemaSpace.__new__(SchemaSpace)
        s.sid = space_id
        s.name = str(space_id)
        cpython.dict.PyDict_SetItem(self.spaces, space_id, s)
        return s

    cdef inline clear(self):
        self.spaces.clear()

    cdef SchemaSpace parse_space(self, space_row):
        cdef:
            SchemaSpace sp
            size_t k
            size_t row_len
            list format_list

        assert space_row is not None

        sp = <SchemaSpace> SchemaSpace.__new__(SchemaSpace)
        row_len = <size_t> len(space_row)

        k = 0
        sp.sid = <int> space_row[k]
        k += 1
        if k < row_len:
            sp.owner = <int> space_row[k]
        k += 1
        if k < row_len:
            sp.name = <str> space_row[k]
        k += 1
        if k < row_len:
            sp.engine = <str> space_row[k]
        k += 1
        if k < row_len:
            sp.field_count = <int> space_row[k]
        k += 1
        if k < row_len:
            sp.flags = space_row[k]
        k += 1
        if k < row_len:
            # format
            format_list = <list> space_row[k]
            if len(format_list) > 0:
                sp.metadata = <Metadata> Metadata.__new__(Metadata)
                for i in range(len(format_list)):
                    field_id = i
                    field = <Field> Field.__new__(Field)

                    field.name = format_list[i]['name']
                    field.type = format_list[i]['type']
                    field.is_nullable = format_list[i].get('is_nullable')

                    sp.metadata.add(field_id, field)

        return sp

    cdef SchemaIndex parse_index(self, index_row):
        cdef:
            SchemaIndex idx
            SchemaSpace sp
            uint32_t i
            int field_id = -1
            str field_type

        assert index_row is not None

        idx = <SchemaIndex> SchemaIndex.__new__(SchemaIndex)
        idx.sid = <int> index_row[0]
        idx.iid = <int> index_row[1]
        idx.name = <str> index_row[2]
        idx.index_type = <str> index_row[3]
        idx.unique = <object> index_row[4]
        idx.metadata = None

        sp = self.get_space(idx.sid)
        if sp is None:
            raise TarantoolSchemaError(
                'Space with id {} not found'.format(idx.sid))

        if sp.metadata is not None:
            parts = index_row[5]
            idx.metadata = <Metadata> Metadata.__new__(Metadata)

            if isinstance(parts, (list, tuple)):
                for part in parts:
                    field = <Field> Field.__new__(Field)
                    field.name = ''

                    if isinstance(part, (list, tuple)):
                        assert len(part) == 2, 'Part len must be 2'
                        field_id = part[0]
                        field.type = part[1]
                    elif isinstance(part, dict):
                        field_id = part['field']
                        field.type = part['type']
                        # TODO: add is_nullable and collation if we really need
                        # TODO: it in a python driver
                    else:
                        raise TypeError(
                            "unexpected type of part: {}. "
                            "must be either dict or list/tuple".format(
                                type(part)
                            )
                        )

                    if field_id < sp.metadata.len():
                        field.name = sp.metadata.name_by_id(field_id)
                    else:
                        logger.debug(
                            'Field #%d of space %s is not '
                            'in space format definition', field_id, sp.name)

                    idx.metadata.add(field_id, field)
            else:
                for i in range(parts):
                    field = <Field> Field.__new__(Field)

                    field_id = index_row[5 + 1 + i * 2]
                    field.type = index_row[5 + 2 + i * 2]
                    field.name = ''

                    if field_id < sp.metadata.len():
                        field.name = sp.metadata.name_by_id(field_id)
                    else:
                        logger.debug(
                            'Field #%d of space %s is not '
                            'in space format definition', field_id, sp.name)

                    idx.metadata.add(field_id, field)

        return idx

    @staticmethod
    cdef Schema parse(int64_t schema_id, spaces, indexes):
        cdef:
            Schema s
            SchemaSpace sp
            SchemaIndex idx

        s = <Schema> Schema.__new__(Schema, <int> schema_id)
        for space_row in spaces:
            sp = s.parse_space(space_row)
            cpython.dict.PyDict_SetItem(s.spaces, sp.sid, sp)
            if sp.name:
                cpython.dict.PyDict_SetItem(s.spaces, sp.name, sp)

        for index_row in indexes:
            idx = s.parse_index(index_row)
            sp = s.spaces[idx.sid]
            sp.add_index(idx)

        return s

    def __repr__(self):  # pragma: nocover
        return '<Schema spaces={}>'.format(len(self.spaces))

cdef list dict_to_list_fields(dict d, Metadata metadata, bint default_none):
    cdef:
        list tpl
        dict used
        object value
        str field_name
        int32_t used_count
        int field_id
        PyObject *obj_p

    assert metadata is not None
    assert d is not None

    tpl = []
    used_count = <int32_t> cpython.dict.PyDict_Size(d)

    for field_id in range(metadata.len()):
        field_name = metadata.name_by_id(field_id)

        obj_p = cpython.dict.PyDict_GetItem(d, field_name)
        if obj_p is not NULL:
            value = <object> obj_p
            used_count -= 1
            tpl.append(value)
        else:
            if default_none:
                tpl.append(None)

    # Warn user if he used any of unknown fields
    if used_count != 0:
        used = {}
        for field_id in range(metadata.len()):
            field_name = metadata.name_by_id(field_id)

            if <bint> cpython.dict.PyDict_Contains(d, field_name):
                used[field_name] = None
        if <bint> cpython.dict.PyDict_Contains(d, ''):
            used[''] = None

        for f in d:
            if f not in used:
                logger.warning(
                    'Field \'%s\' in supplied dict is unknown as '
                    'a tuple field for selected index. Skipping.',
                    f
                )
    return tpl



import asyncio

import asynctnt
from asynctnt.exceptions import TarantoolDatabaseError, TarantoolNotConnectedError
from tests import BaseTarantoolTestCase
from tests.util import get_big_param, get_complex_param


class CommonTestCase(BaseTarantoolTestCase):
    async def test__encoding_utf8(self):
        p, p_cmp = get_complex_param(replace_bin=False)

        data = [1, "hello", 1, 0, p]
        data_cmp = [1, "hello", 1, 0, p_cmp]

        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertResponseEqual(res, [data_cmp], "Body ok")

        res = await self.conn.select(self.TESTER_SPACE_ID)
        self.assertResponseEqual(res, [data_cmp], "Body ok")

    async def test__encoding_cp1251(self):
        await self.tnt_reconnect(encoding="cp1251")
        p, p_cmp = get_complex_param(replace_bin=False)

        data = [1, "hello", 1, 0, p]
        data_cmp = [1, "hello", 1, 0, p_cmp]

        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertResponseEqual(res, [data_cmp], "Body ok")

        res = await self.conn.select(self.TESTER_SPACE_ID)
        self.assertResponseEqual(res, [data_cmp], "Body ok")

    async def test__schema_refetch_on_schema_change(self):
        await self.tnt_reconnect(auto_refetch_schema=True, username="t1", password="t1")
        self.assertTrue(self.conn.fetch_schema)
        self.assertTrue(self.conn.auto_refetch_schema)
        schema_before = self.conn.schema_id
        self.assertNotEqual(schema_before, -1)

        # Changing scheme
        await self.conn.eval("box.schema.create_space('new_space');")

        try:
            try:
                await self.conn.ping()
            except Exception as e:
                self.fail(e)

            # wait for schema to refetch
            await self.sleep(1)

            self.assertGreater(self.conn.schema_id, schema_before, "Schema changed")
            self.assertIn("new_space", self.conn.schema.spaces)
        finally:
            await self.conn.eval(
                "local s = box.space.new_space;" "if s ~= nil then s:drop(); end"
            )

    async def test__schema_refetch_manual(self):
        await self.tnt_reconnect(
            fetch_schema=True, auto_refetch_schema=False, username="t1", password="t1"
        )
        self.assertTrue(self.conn.fetch_schema)
        self.assertFalse(self.conn.auto_refetch_schema)
        schema_before = self.conn.schema_id
        self.assertNotEqual(schema_before, -1)

        await self.conn.call("change_format")

        try:
            await self.conn.ping()
        except Exception as e:
            self.fail(e)

        self.assertEqual(self.conn.schema_id, schema_before, "schema not changed")

        await self.conn.refetch_schema()

        self.assertGreater(self.conn.schema_id, schema_before, "Schema changed")
        sp = self.conn.schema.spaces[self.TESTER_SPACE_NAME]
        self.assertEqual(6, len(sp.metadata.fields))
        self.assertEqual("f6", sp.metadata.fields[5].name)
        self.assertEqual("*", sp.metadata.fields[5].type)

    async def test__schema_no_fetch_and_refetch(self):
        await self.tnt_reconnect(
            auto_refetch_schema=False, username="t1", password="t1", fetch_schema=False
        )
        self.assertFalse(self.conn.fetch_schema)
        self.assertFalse(self.conn.auto_refetch_schema)
        self.assertEqual(self.conn.schema_id, -1)

        # Changing scheme
        await self.conn.eval("s = box.schema.create_space('new_space');" "s:drop();")

        try:
            await self.conn.ping()
        except Exception as e:
            self.fail(e)

        await asyncio.sleep(1)  # wait for potential schema refetch

        self.assertEqual(self.conn.schema_id, -1)

    async def test__parse_numeric_map_keys(self):
        res = await self.conn.eval(
            """return {
                [1] = 1,
                [2] = 2,
                hello = 3,
                world = 4,
                [-3] = 5,
                [4.5] = 6
            }"""
        )

        d = {1: 1, 2: 2, "hello": 3, "world": 4, -3: 5, 4.5: 6}

        self.assertDictEqual(res[0], d, "Numeric keys parsed ok")

    async def test__read_buffer_reallocate_ok(self):
        await self.tnt_reconnect(initial_read_buffer_size=1)

        p, cmp = get_complex_param(
            encoding=self.conn.encoding, replace_bin=self.conn.version < (3, 0)
        )
        try:
            res = await self.conn.call("func_param", [p])
        except Exception as e:
            self.fail(e)

        self.assertDictEqual(res[0][0], cmp, "Body ok")

    async def test__read_buffer_deallocate_ok(self):
        size = 100 * 1000
        await self.tnt_reconnect(initial_read_buffer_size=size)

        # Waiting big response, so ReadBuffer grows to hold it
        p = get_big_param(size=size * 3)
        try:
            await self.conn.call("func_param", [p])
        except Exception as e:
            self.fail(e)

        # Waiting small response, so ReadBuffer deallocates memory
        p = get_big_param(size=10)
        try:
            await self.conn.call("func_param", [p])
        except Exception as e:
            self.fail(e)

    async def test__write_buffer_reallocate(self):
        p = get_big_param(size=100 * 1024)
        try:
            res = await self.conn.call("func_param", [p])
        except Exception as e:
            self.fail(e)

        self.assertDictEqual(res[0][0], p, "Body ok")

    async def test__ensure_no_attribute_error_on_not_connected(self):
        await self.tnt_disconnect()

        self._conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port)

        with self.assertRaises(TarantoolNotConnectedError):
            await self.conn.ping()

    async def test__encode_unsupported_type(self):
        class A:
            pass

        with self.assertRaisesRegex(
            TypeError, "Type `(.+)` is not supported for encoding"
        ):
            await self.conn.call("func_param", [{"a": A()}])

    async def test__schema_refetch_next_byte(self):
        await self.tnt_reconnect(auto_refetch_schema=True, username="t1", password="t1")
        await self.conn.call("func_hello")

        # Changing scheme
        try:
            for _ in range(251):
                await self.conn.eval(
                    "s = box.schema.create_space('new_space');" "s:drop();"
                )
        except TarantoolDatabaseError as e:
            self.fail(e)

        try:
            for _ in range(1, 255):
                await self.conn.call("func_hello")
        except TarantoolDatabaseError as e:
            self.fail(e)

    async def test__schema_refetch_unknown_space(self):
        await self.tnt_reconnect(
            auto_refetch_schema=True, username="t1", password="t1", ping_timeout=0.1
        )

        async def func():
            # trying to select from an unknown space until it is created
            while True:
                try:
                    await self.conn.select("spacex")
                    return
                except Exception:
                    pass

                await asyncio.sleep(0.1)

        f = asyncio.ensure_future(asyncio.wait_for(func(), timeout=1))

        # Changing scheme
        try:
            conn = await asynctnt.connect(
                host=self.tnt.host, port=self.tnt.port, username="t1", password="t1"
            )
            async with conn:
                await conn.eval(
                    "s = box.schema.create_space('spacex');"
                    "s:create_index('primary');"
                )
        except TarantoolDatabaseError as e:
            self.fail(e)

        try:
            await f
        except (asyncio.TimeoutError, asyncio.CancelledError) as e:
            self.fail("Schema is not updated: %s %s" % (type(e), e))

    async def test__schema_refetch_on_disconnect_race_condition(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, username="t1", password="t1"
        )
        await conn.connect()
        await conn.eval("require('msgpack').cfg{encode_use_tostring = true}")
        await conn.call("box.schema.space.create", ["geo", {"if_not_exists": True}])
        await conn.call("box.space.geo:format", [[{"name": "id", "type": "string"}]])
        await conn.disconnect()



cimport cython
from cpython.mem cimport PyMem_Free, PyMem_Malloc, PyMem_Realloc
from libc.string cimport memcpy, memmove


@cython.no_gc_clear
@cython.final
cdef class ReadBuffer:
    def __cinit__(self):
        self.buf = NULL
        self.initial_buffer_size = 0
        self.len = 0
        self.use = 0
        self.encoding = None

    @staticmethod
    cdef ReadBuffer create(str encoding, size_t initial_buffer_size=0x80000):
        cdef ReadBuffer b
        b = ReadBuffer.__new__(ReadBuffer)

        b.buf = <char*>PyMem_Malloc(sizeof(char) * <size_t>initial_buffer_size)
        if b.buf is NULL:
            raise MemoryError

        b.initial_buffer_size = initial_buffer_size
        b.len = initial_buffer_size
        b.use = 0
        b.encoding = encoding
        return b

    def __dealloc__(self):
        if self.buf is not NULL:
            PyMem_Free(self.buf)
            self.buf = NULL
        self.initial_buffer_size = 0
        self.len = 0
        self.use = 0

    cdef void _reallocate(self, size_t new_size) except *:
        cdef char *new_buf

        # print('ReadBuffer reallocate: {}'.format(new_size))
        new_buf = <char*>PyMem_Realloc(<void*>self.buf, <size_t>new_size)
        if new_buf is NULL:
            PyMem_Free(self.buf)
            self.buf = NULL
            self.initial_buffer_size = 0
            self.len = 0
            self.use = 0
            raise MemoryError
        self.buf = new_buf
        self.len = new_size

    cdef int extend(self, const char *data, size_t len) except -1:
        cdef:
            size_t new_size
            size_t dealloc_threshold
        new_size = self.use + len
        dealloc_threshold = self.len // _DEALLOCATE_RATIO
        if new_size > self.len:
            self._reallocate(
                size_t_max(nearest_power_of_2(new_size), self.len << 1)
            )
        elif dealloc_threshold >= self.initial_buffer_size \
                and new_size < dealloc_threshold:
            self._reallocate(dealloc_threshold)

        memcpy(&self.buf[self.use], data, len)
        self.use += len
        return 0

    cdef void move(self, size_t pos):
        cdef size_t delta = self.use - pos
        memmove(self.buf, &self.buf[pos], delta)
        self.use = delta

    cdef void move_offset(self, ssize_t offset, size_t size) except *:
        cdef size_t dealloc_threshold = self.len // _DEALLOCATE_RATIO
        if offset == 0:
            return
        assert offset > 0, \
            'Offset incorrect. Got: {}. use:{}, len:{}'.format(
                offset, self.use, self.len
            )
        memmove(self.buf, &self.buf[offset], size)

        if dealloc_threshold >= self.initial_buffer_size \
                and size < dealloc_threshold:
            self._reallocate(dealloc_threshold)

    cdef bytes get_slice(self, size_t begin, size_t end):
        cdef:
            ssize_t diff
            char *p
        p = &self.buf[begin]
        diff = end - begin
        return <bytes>p[:diff]

    cdef bytes get_slice_begin(self, size_t begin):
        cdef:
            ssize_t diff
            char *p
        p = &self.buf[begin]
        diff = self.use - begin
        return <bytes>p[:diff]

    cdef bytes get_slice_end(self, size_t end):
        cdef:
            ssize_t diff
            char *p
        p = &self.buf[0]
        diff = end - 0
        return <bytes>p[:diff]



cimport cpython


cdef extern from "tupleobj/tupleobj.h":

    cpython.PyTypeObject *AtntTuple_InitTypes() except NULL

    int AtntTuple_CheckExact(object)
    object AtntTuple_New(object, int)
    void AtntTuple_SET_ITEM(object, int, object)

    object AtntTupleDesc_New(object, object)



from decimal import Decimal

from asynctnt import Response
from asynctnt.exceptions import ErrorCode, TarantoolDatabaseError, TarantoolSchemaError
from tests import BaseTarantoolTestCase
from tests._testbase import ensure_version


class UpdateTestCase(BaseTarantoolTestCase):
    async def _fill_data(self):
        data = [
            [0, "a", 1, 5, "data1"],
            [1, "b", 8, 6, "data2"],
            [2, "c", 10, 12, "data3", "extra_field"],
            [3, "d", 14, 16, "data4", Decimal("12.3"), 12.5],
            [
                4,
                "e",
                18,
                20,
                {
                    "tree1": {
                        "tree11": {
                            "key1": "value1",
                        }
                    }
                },
            ],
        ]
        for t in data:
            await self.conn.insert(self.TESTER_SPACE_ID, t)

        return data

    async def test__update_one_assign(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["=", 2, 2]])
        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")

        data[1][2] = 2
        self.assertResponseEqual(res, [data[1]], "Body ok")

    @ensure_version(min=(2, 3))
    async def test__update_one_assign_by_field_name_with_no_schema(self):
        data = await self._fill_data()

        await self.tnt_reconnect(fetch_schema=False)

        res = await self.conn.update(self.TESTER_SPACE_ID, [4], [["=", "f4", 100]])
        data[4][3] = 100
        self.assertResponseEqual(res, [data[4]], "Body ok")

    @ensure_version(min=(2, 3))
    async def test__update_one_assign_by_json(self):
        data = await self._fill_data()

        res = await self.conn.update(
            self.TESTER_SPACE_ID, [4], [["=", "f5.tree1.tree11.key1", "value2"]]
        )
        data[4][4]["tree1"]["tree11"]["key1"] = "value2"
        self.assertResponseEqual(res, [data[4]], "Body ok")

    async def test__update_one_insert(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["!", 2, 14]])
        data[1].insert(2, 14)
        self.assertResponseEqual(res, [data[1]], "Body ok")

    async def test__update_one_delete(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [2], [["#", 5, 1]])
        data[2].pop(5)
        self.assertResponseEqual(res, [data[2]], "Body ok")

    async def test__update_one_plus(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["+", 2, 3]])
        data[1][2] += 3
        self.assertResponseEqual(res, [data[1]], "Body ok")

    @ensure_version(min=(2, 3))
    async def test__update_one_plus_decimal(self):
        data = await self._fill_data()

        add = Decimal("1.1")
        res = await self.conn.update(self.TESTER_SPACE_ID, [3], [["+", 5, add]])
        data[3][5] += add
        self.assertResponseEqual(res, [data[3]], "Body ok")

    @ensure_version(min=(2, 3))
    async def test__update_one_plus_float(self):
        data = await self._fill_data()

        add = 1.5
        res = await self.conn.update(self.TESTER_SPACE_ID, [3], [["+", 6, add]])
        data[3][6] += add
        self.assertResponseEqual(res, [data[3]], "Body ok")

    async def test__update_one_plus_str_field(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["+", "f3", 3]])
        data[1][2] += 3
        self.assertResponseEqual(res, [data[1]], "Body ok")

    @ensure_version(min=(2, 3))
    async def test__update_one_plus_str_field_unknown(self):
        await self._fill_data()

        with self.assertRaisesRegex(
            TarantoolDatabaseError, "Field 'f10' was not found in the tuple"
        ):
            await self.conn.update(self.TESTER_SPACE_ID, [1], [["+", "f10", 3]])

    async def test__update_one_plus_negative(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["+", 2, -3]])
        data[1][2] += -3
        self.assertResponseEqual(res, [data[1]], "Body ok")

    async def test__update_one_minus(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [0], [["-", 2, 1]])
        data[0][2] -= 1
        self.assertResponseEqual(res, [data[0]], "Body ok")

    async def test__update_one_minus_negative(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["-", 2, -3]])
        data[1][2] -= -3
        self.assertResponseEqual(res, [data[1]], "Body ok")

    async def test__update_one_band(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["&", 2, 3]])
        data[1][2] &= 3
        self.assertResponseEqual(res, [data[1]], "Body ok")

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["&", 2, 2]])
        data[1][2] &= 2
        self.assertResponseEqual(res, [data[1]], "Body ok")

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["&", 2, 1]])
        data[1][2] &= 1
        self.assertResponseEqual(res, [data[1]], "Body ok")

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["&", 2, 0]])
        data[1][2] &= 0
        self.assertResponseEqual(res, [data[1]], "Body ok")

    async def test__update_one_bor(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["|", 2, 3]])
        data[1][2] |= 3
        self.assertResponseEqual(res, [data[1]], "Body ok")

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["|", 2, 2]])
        data[1][2] |= 2
        self.assertResponseEqual(res, [data[1]], "Body ok")

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["|", 2, 1]])
        data[1][2] |= 1
        self.assertResponseEqual(res, [data[1]], "Body ok")

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["|", 2, 0]])
        data[1][2] |= 0
        self.assertResponseEqual(res, [data[1]], "Body ok")

    async def test__update_one_bxor(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["^", 2, 3]])
        data[1][2] ^= 3
        self.assertResponseEqual(res, [data[1]], "Body ok")

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["^", 2, 2]])
        data[1][2] ^= 2
        self.assertResponseEqual(res, [data[1]], "Body ok")

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["^", 2, 1]])
        data[1][2] ^= 1
        self.assertResponseEqual(res, [data[1]], "Body ok")

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["^", 2, 0]])
        data[1][2] ^= 0
        self.assertResponseEqual(res, [data[1]], "Body ok")

    @ensure_version(min=(2, 3))
    async def test__update_operations_not_int_without_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        data = [1, "hello2", 1, 4, "what is up"]
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [["+", "f3", 1]])
        data[2] += 1
        self.assertResponseEqual(res, [data], "Body ok")

    async def test__update_splice(self):
        data = [1, "hello2", 1, 4, "what is up"]
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], [[":", 1, 1, 3, "!!!"]])

        data[1] = "h!!!o2"
        self.assertResponseEqual(res, [data], "Body ok")

    async def test__update_splice_bytes(self):
        data = [1, "hello2", 1, 4, "what is up", -5]
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.update(
            self.TESTER_SPACE_ID, [1], [[b":", 1, 1, 3, "!!!"]]
        )

        data[1] = "h!!!o2"
        self.assertResponseEqual(res, [data], "Body ok")

    async def test__update_splice_wrong_args(self):
        data = [1, "hello2", 1, 4, "what is up"]
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        with self.assertRaisesRegex(IndexError, r"Operation length must be at least 3"):
            await self.conn.update(self.TESTER_SPACE_ID, [1], [[":", 2]])

        with self.assertRaisesRegex(
            ValueError, r"Splice operation must have length of 5"
        ):
            await self.conn.update(self.TESTER_SPACE_ID, [1], [[":", 2, 1]])

        with self.assertRaisesRegex(
            ValueError, r"Splice operation must have length of 5"
        ):
            await self.conn.update(self.TESTER_SPACE_ID, [1], [[":", 2, 1, 3]])

        with self.assertRaisesRegex(TypeError, r"Splice offset must be int"):
            await self.conn.update(self.TESTER_SPACE_ID, [1], [[":", 2, 1, {}, ":::"]])

        with self.assertRaisesRegex(TypeError, r"Splice position must be int"):
            await self.conn.update(self.TESTER_SPACE_ID, [1], [[":", 2, {}, {}, ":::"]])

        with self.assertRaisesRegex(
            TypeError, r"Operation field_no must be " r"of either int or str type"
        ):
            await self.conn.update(
                self.TESTER_SPACE_ID, [1], [[":", {}, {}, {}, ":::"]]
            )

        with self.assertRaisesRegex(TypeError, r"Unknown update operation type `yo`"):
            await self.conn.update(self.TESTER_SPACE_ID, [1], [["yo", 1, 2, 3]])

        with self.assertRaisesRegex(
            TypeError, r"Operation type must of a str or bytes type"
        ):
            await self.conn.update(self.TESTER_SPACE_ID, [1], [[{}, 1, 2, 3]])

        with self.assertRaisesRegex(
            TypeError, r"Single operation must be a tuple or list"
        ):
            await self.conn.update(self.TESTER_SPACE_ID, [1], [{}])

        with self.assertRaises(TarantoolDatabaseError) as exc:
            await self.conn.update(self.TESTER_SPACE_ID, [1], [("+", 2, {})])
        self.assertRegex(
            exc.exception.message,
            r"Argument type in operation '\+' on field \d does not match field type: expected a number",
        )

    async def test__update_multiple_operations(self):
        t = [1, "1", 1, 5, "hello", 3, 4, 8]
        await self.conn.insert(self.TESTER_SPACE_ID, t)

        t[2] += 1
        t[3] -= 4
        t[5] &= 5
        t[6] |= 7
        t[7] = 100
        t[4] = "h!!!o"

        operations = [
            ["+", 2, 1],
            ["-", 3, 4],
            ["&", 5, 5],
            ["|", 6, 7],
            ["=", 7, 100],
            [":", 4, 1, 3, "!!!"],
        ]

        res = await self.conn.update(self.TESTER_SPACE_ID, [1], operations)
        self.assertResponseEqual(res, [t], "Body ok")

    async def test__update_by_name(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_NAME, [1], [["=", 2, 2]])

        data[1][2] = 2
        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [data[1]], "Body ok")

    async def test__update_by_name_no_schema(self):
        await self._fill_data()

        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.update(self.TESTER_SPACE_NAME, [1], [["=", 2, 2]])

    async def test__update_by_index_id(self):
        index_name = "temp_idx"
        res = self.tnt.command('make_third_index("{}")'.format(index_name))
        index_id = res[0][0]

        try:
            await self.tnt_reconnect()
            data = await self._fill_data()

            res = await self.conn.update(
                self.TESTER_SPACE_ID, [data[0][2]], [("=", 2, 1)], index=index_id
            )

            data[0][2] = 1
            self.assertIsInstance(res, Response, "Got response")
            self.assertEqual(res.code, 0, "success")
            self.assertGreater(res.sync, 0, "sync > 0")
            self.assertResponseEqual(res, [data[0]], "Body ok")
        finally:
            self.tnt.command(
                "box.space.{}.index.{}:drop()".format(
                    self.TESTER_SPACE_NAME, index_name
                )
            )

    async def test__select_by_index_name(self):
        index_name = "temp_idx"
        res = self.tnt.command('make_third_index("{}")'.format(index_name))

        try:
            await self.tnt_reconnect()
            data = await self._fill_data()

            res = await self.conn.update(
                self.TESTER_SPACE_ID, [data[0][2]], [["=", 2, 1]], index=index_name
            )

            data[0][2] = 1
            self.assertIsInstance(res, Response, "Got response")
            self.assertEqual(res.code, 0, "success")
            self.assertGreater(res.sync, 0, "sync > 0")
            self.assertResponseEqual(res, [data[0]], "Body ok")
        finally:
            self.tnt.command(
                "box.space.{}.index.{}:drop()".format(
                    self.TESTER_SPACE_NAME, index_name
                )
            )

    async def test__update_by_index_id_no_schema(self):
        await self._fill_data()
        await self.tnt_reconnect(fetch_schema=False)

        try:
            await self.conn.update(self.TESTER_SPACE_ID, [0], [["=", 2, 1]], index=0)
        except Exception as e:
            self.fail(e)

    async def test__update_by_index_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.update(
                self.TESTER_SPACE_NAME, [0], [["=", 2, 1]], index="primary"
            )

    async def test__update_operations_none(self):
        data = await self._fill_data()
        try:
            res = await self.conn.update(self.TESTER_SPACE_NAME, [data[0][0]], None)
        except TarantoolDatabaseError as e:
            if self.conn.version < (1, 7):
                if e.code == ErrorCode.ER_ILLEGAL_PARAMS:
                    # success
                    return
            raise
        self.assertResponseEqual(res, [data[0]], "empty operations")

    async def test__update_dict_key(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, {"f1": 0}, [["+", 3, 1]])
        data[0][3] += 1
        self.assertResponseEqual(res, [data[0]], "Body ok")

    async def test__update_dict_resp(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID, [0], [["+", 3, 1]])
        data[0][3] += 1

        self.assertResponseEqualKV(
            res,
            [
                {
                    "f1": data[0][0],
                    "f2": data[0][1],
                    "f3": data[0][2],
                    "f4": data[0][3],
                    "f5": data[0][4],
                }
            ],
        )



cimport cython


@cython.final
cdef class BeginRequest(BaseRequest):
    cdef int encode_body(self, WriteBuffer buffer) except -1:
        cdef:
            char *p
            char *begin
            uint32_t body_map_sz
            uint32_t max_body_len

        body_map_sz = 2
        # Size description:
        max_body_len = mp_sizeof_map(body_map_sz) \
                       + mp_sizeof_uint(tarantool.IPROTO_TXN_ISOLATION) \
                       + mp_sizeof_uint(self.isolation) \
                       + mp_sizeof_uint(tarantool.IPROTO_TIMEOUT) \
                       + mp_sizeof_double(self.tx_timeout)

        buffer.ensure_allocated(max_body_len)

        p = begin = &buffer._buf[buffer._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_TXN_ISOLATION)
        p = mp_encode_uint(p, self.isolation)
        p = mp_encode_uint(p, tarantool.IPROTO_TIMEOUT)
        p = mp_encode_double(p, self.tx_timeout)

        buffer._length += (p - begin)

@cython.final
cdef class CommitRequest(BaseRequest):
    pass

@cython.final
cdef class RollbackRequest(BaseRequest):
    pass



cimport cython


@cython.final
cdef class SelectRequest(BaseRequest):
    cdef int encode_body(self, WriteBuffer buffer) except -1:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len
            uint32_t space_id, index_id

        space_id = self.space.sid
        index_id = self.index.iid

        body_map_sz = 3 \
                      + <uint32_t> (index_id > 0) \
                      + <uint32_t> (self.offset > 0) \
                      + <uint32_t> (self.iterator > 0)
        # Size description:
        # mp_sizeof_map(body_map_sz)
        # + mp_sizeof_uint(TP_SPACE)
        # + mp_sizeof_uint(space)
        # + mp_sizeof_uint(TP_LIMIT)
        # + mp_sizeof_uint(limit)
        max_body_len = 1 \
                       + 1 \
                       + 9 \
                       + 1 \
                       + 9

        if index_id > 0:
            # mp_sizeof_uint(TP_INDEX) + mp_sizeof_uint(index_id)
            max_body_len += 1 + 9
        if self.offset > 0:
            # mp_sizeof_uint(TP_OFFSET) + mp_sizeof_uint(offset)
            max_body_len += 1 + 9
        if self.iterator > 0:
            # mp_sizeof_uint(TP_ITERATOR) + mp_sizeof_uint(iterator)
            max_body_len += 1 + 1

        max_body_len += 1  # mp_sizeof_uint(TP_KEY);

        buffer.ensure_allocated(max_body_len)

        p = begin = &buffer._buf[buffer._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
        p = mp_encode_uint(p, space_id)
        p = mp_encode_uint(p, tarantool.IPROTO_LIMIT)
        p = mp_encode_uint(p, self.limit)

        if index_id > 0:
            p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
            p = mp_encode_uint(p, index_id)
        if self.offset > 0:
            p = mp_encode_uint(p, tarantool.IPROTO_OFFSET)
            p = mp_encode_uint(p, self.offset)
        if self.iterator > 0:
            p = mp_encode_uint(p, tarantool.IPROTO_ITERATOR)
            p = mp_encode_uint(p, self.iterator)

        p = mp_encode_uint(p, tarantool.IPROTO_KEY)
        buffer._length += (p - begin)
        p = encode_key_sequence(buffer, p, self.key, self.index.metadata, False)



cimport cython


@cython.final
cdef class Db:
    def __cinit__(self):
        self._stream_id = 0
        self._protocol = None
        self._encoding = None

    @staticmethod
    cdef inline Db create(BaseProtocol protocol, uint64_t stream_id):
        cdef Db db = Db.__new__(Db)
        db._stream_id = stream_id
        db._protocol = protocol
        db._encoding = protocol.encoding
        return db

    cdef inline uint64_t next_sync(self):
        return self._protocol.next_sync()

    cdef object _ping(self, float timeout):
        cdef PingRequest req = PingRequest.__new__(PingRequest)
        req.op = tarantool.IPROTO_PING
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.check_schema_change = True
        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _id(self, float timeout):
        cdef IDRequest req = IDRequest.__new__(IDRequest)
        req.op = tarantool.IPROTO_ID
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.check_schema_change = False
        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _auth(self,
                      bytes salt,
                      str username,
                      str password,
                      float timeout):
        cdef:
            AuthRequest req

        req = AuthRequest.__new__(AuthRequest)
        req.op = tarantool.IPROTO_AUTH
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.salt = salt
        req.username = username
        req.password = password
        req.push_subscribe = False
        req.parse_as_tuples = False
        req.parse_metadata = False
        req.check_schema_change = False

        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _call(self,
                      tarantool.iproto_type op,
                      str func_name,
                      object args,
                      float timeout,
                      bint push_subscribe):
        cdef CallRequest req = CallRequest.__new__(CallRequest)
        req.op = op
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.func_name = func_name
        req.args = args
        req.push_subscribe = push_subscribe
        req.check_schema_change = True
        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _eval(self,
                      str expression,
                      object args,
                      float timeout,
                      bint push_subscribe):
        cdef EvalRequest req = EvalRequest.__new__(EvalRequest)
        req.op = tarantool.IPROTO_EVAL
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.expression = expression
        req.args = args
        req.push_subscribe = push_subscribe
        req.check_schema_change = True
        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _select(self,
                        object space,
                        object index,
                        object key,
                        uint64_t offset,
                        uint64_t limit,
                        object iterator,
                        float timeout,
                        bint check_schema_change):
        cdef:
            SchemaSpace sp
            SchemaIndex idx
            SelectRequest req
            uint32_t iterator_value

        sp = self._protocol._schema.get_or_create_space(space)
        idx = sp.get_index(index)

        iterator_value = self._protocol.transform_iterator(iterator)
        if key is None and isinstance(iterator, int):
            iterator_value = <uint32_t> iterator
            if iterator_value == 0:
                iterator_value = 2  # ALL

        req = SelectRequest.__new__(SelectRequest)
        req.op = tarantool.IPROTO_SELECT
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.space = sp
        req.index = idx
        req.key = key
        req.offset = offset
        req.limit = limit
        req.iterator = iterator_value
        req.push_subscribe = False
        req.check_schema_change = check_schema_change
        req.parse_as_tuples = True

        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _insert(self,
                        object space,
                        object t,
                        bint replace,
                        float timeout):
        cdef:
            SchemaSpace sp
            InsertRequest req

        sp = self._protocol._schema.get_or_create_space(space)

        req = InsertRequest.__new__(InsertRequest)
        req.op = tarantool.IPROTO_REPLACE if replace else tarantool.IPROTO_INSERT
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.space = sp
        req.t = t
        req.push_subscribe = False
        req.check_schema_change = True
        req.parse_as_tuples = True

        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _delete(self,
                        object space,
                        object index,
                        object key,
                        float timeout):
        cdef:
            SchemaSpace sp
            SchemaIndex idx
            DeleteRequest req

        sp = self._protocol._schema.get_or_create_space(space)
        idx = sp.get_index(index)

        req = DeleteRequest.__new__(DeleteRequest)
        req.op = tarantool.IPROTO_DELETE
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.space = sp
        req.index = idx
        req.key = key
        req.push_subscribe = False
        req.check_schema_change = True
        req.parse_as_tuples = True

        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _update(self,
                        object space,
                        object index,
                        object key,
                        list operations,
                        float timeout):
        cdef:
            SchemaSpace sp
            SchemaIndex idx
            UpdateRequest req

        sp = self._protocol._schema.get_or_create_space(space)
        idx = sp.get_index(index)

        req = UpdateRequest.__new__(UpdateRequest)
        req.op = tarantool.IPROTO_UPDATE
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.space = sp
        req.index = idx
        req.key = key
        req.operations = operations
        req.push_subscribe = False
        req.check_schema_change = True
        req.parse_as_tuples = True

        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _upsert(self,
                        object space,
                        object t,
                        list operations,
                        float timeout):
        cdef:
            SchemaSpace sp
            UpsertRequest req

        sp = self._protocol._schema.get_or_create_space(space)

        req = UpsertRequest.__new__(UpsertRequest)
        req.op = tarantool.IPROTO_UPSERT
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.space = sp
        req.t = t
        req.operations = operations
        req.push_subscribe = False
        req.check_schema_change = True
        req.parse_as_tuples = True

        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _execute(self,
                         object query,
                         object args,
                         bint parse_metadata,
                         float timeout):
        cdef:
            ExecuteRequest req

        req = ExecuteRequest.__new__(ExecuteRequest)
        req.op = tarantool.IPROTO_EXECUTE
        req.sync = self.next_sync()
        req.stream_id = self._stream_id

        if isinstance(query, str):
            req.query = query
            req.statement_id = 0
        elif isinstance(query, int):
            req.query = None
            req.statement_id = <uint64_t> query
        else:
            raise TypeError('query must be either str or int')

        req.args = args
        req.parse_metadata = parse_metadata
        req.push_subscribe = False
        req.check_schema_change = True
        req.parse_as_tuples = True

        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _prepare(self,
                         object query,
                         bint parse_metadata,
                         float timeout):
        cdef:
            PrepareRequest req

        req = PrepareRequest.__new__(PrepareRequest)
        req.op = tarantool.IPROTO_PREPARE
        req.sync = self.next_sync()
        req.stream_id = self._stream_id

        if isinstance(query, str):
            req.query = query
            req.statement_id = 0
        elif isinstance(query, int):
            req.query = None
            req.statement_id = <uint64_t> query
        else:
            raise TypeError('query must be either str or int')

        req.push_subscribe = False
        req.check_schema_change = True
        req.parse_as_tuples = True
        req.parse_metadata = parse_metadata

        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _begin(self,
                       uint32_t isolation,
                       double tx_timeout,
                       float timeout):
        cdef BeginRequest req = BeginRequest.__new__(BeginRequest)
        req.op = tarantool.IPROTO_BEGIN
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.check_schema_change = True
        req.push_subscribe = False
        req.isolation = isolation
        req.tx_timeout = tx_timeout
        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _commit(self, float timeout):
        cdef CommitRequest req = CommitRequest.__new__(CommitRequest)
        req.op = tarantool.IPROTO_COMMIT
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.check_schema_change = True
        req.push_subscribe = False
        return self._protocol.execute(self._protocol, req, timeout)

    cdef object _rollback(self, float timeout):
        cdef RollbackRequest req = RollbackRequest.__new__(RollbackRequest)
        req.op = tarantool.IPROTO_ROLLBACK
        req.sync = self.next_sync()
        req.stream_id = self._stream_id
        req.check_schema_change = True
        return self._protocol.execute(self._protocol, req, timeout)

    # public methods

    @property
    def stream_id(self):
        return <int> self._stream_id

    def set_stream_id(self, int stream_id):
        self._stream_id = <uint64_t> stream_id

    def ping(self, float timeout=-1):
        return self._ping(timeout)

    def call16(self,
               str func_name,
               object args=None,
               float timeout=-1,
               bint push_subscribe=False):
        return self._call(tarantool.IPROTO_CALL_16,
                          func_name,
                          args,
                          timeout,
                          <bint> push_subscribe)

    def call(self,
             str func_name,
             object args=None,
             float timeout=-1,
             bint push_subscribe=False):
        return self._call(tarantool.IPROTO_CALL,
                          func_name,
                          args,
                          timeout,
                          <bint> push_subscribe)

    def eval(self,
             str expression,
             object args=None,
             float timeout=-1,
             bint push_subscribe=False):
        return self._eval(expression,
                          args,
                          timeout,
                          <bint> push_subscribe)

    def select(self,
               object space,
               object key=None,
               int offset=0,
               int limit=0xffffffff,
               object index=0,
               object iterator=0,
               float timeout=-1,
               bint check_schema_change=True):
        return self._select(space, index, key, offset, limit, iterator,
                            timeout, check_schema_change)

    def insert(self,
               object space,
               object t,
               bint replace=False,
               float timeout=-1):
        return self._insert(space, t, <bint> replace, timeout)

    def replace(self,
                object space,
                object t,
                float timeout=-1):
        return self._insert(space, t, <bint> True, timeout)

    def delete(self,
               object space,
               object key,
               object index=0,
               float timeout=-1):
        return self._delete(space, index, key, timeout)

    def update(self,
               object space,
               object key,
               list operations,
               object index=0,
               float timeout=-1):
        return self._update(space, index, key, operations, timeout)

    def upsert(self,
               object space,
               object t,
               list operations,
               float timeout=-1):
        return self._upsert(space, t, operations, timeout)

    def execute(self,
                object query,
                object args,
                bint parse_metadata=True,
                float timeout=-1):
        return self._execute(query, args, <bint> parse_metadata, timeout)

    def prepare(self,
                object query,
                bint parse_metadata=True,
                float timeout=-1):
        return self._prepare(query, <bint> parse_metadata, timeout)

    def begin(self,
              uint32_t isolation,
              float tx_timeout=0,
              float timeout=-1):
        return self._begin(isolation, <double> tx_timeout, timeout)

    def commit(self, float timeout=-1):
        return self._commit(timeout)

    def rollback(self, float timeout=-1):
        return self._rollback(timeout)



import os
import warnings

from asynctnt import Response
from asynctnt.exceptions import TarantoolNotConnectedError
from tests import BaseTarantoolTestCase


class PingTestCase(BaseTarantoolTestCase):
    SMALL_TIMEOUT = 0.00000000001

    async def test__ping_basic(self):
        res = await self.conn.ping()
        self.assertIsNotNone(res)
        self.assertIsInstance(res, Response)
        self.assertGreater(res.sync, 0, "Sync is not 0")
        self.assertEqual(res.code, 0, "Code is 0")
        self.assertEqual(res.return_code, 0, "Return code is 0")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            self.assertIsNone(res.body, "No body for ping")

    async def test__ping_timeout_on_conn(self):
        await self.tnt_reconnect(request_timeout=self.SMALL_TIMEOUT)
        self.assertEqual(self.conn.request_timeout, self.SMALL_TIMEOUT)

        try:
            await self.conn.ping(timeout=1)
        except Exception as e:
            self.fail("Should not fail on timeout 1: {}".format(e))

    async def test__ping_connection_lost(self):
        self.tnt.stop()
        await self.sleep(0)

        try:
            os.kill(self.tnt.pid, 0)
            running = True
        except Exception:
            running = False

        with self.assertRaises(TarantoolNotConnectedError):
            res = await self.conn.ping()
            print(res)
            print("running", running)
            print(os.system("ps aux | grep tarantool"))

        self.tnt.start()
        await self.sleep(1)

        try:
            await self.conn.ping()
        except Exception as e:
            self.fail("Should not fail on timeout 1: {}".format(e))

    async def test__ping_with_reconnect(self):
        await self.conn.reconnect()
        res = await self.conn.ping()
        self.assertIsInstance(res, Response, "Ping result")



cimport cython


@cython.final
cdef class UpsertRequest(BaseRequest):
    cdef int encode_body(self, WriteBuffer buffer) except -1:
        return encode_request_update(buffer, self.space, self.space.get_index(0),
                                     self.t, self.operations, <bint> True)



from asynctnt import Response
from asynctnt.exceptions import TarantoolSchemaError
from tests import BaseTarantoolTestCase
from tests.util import get_complex_param


class InsertTestCase(BaseTarantoolTestCase):
    async def test__insert_one(self):
        data = [1, "hello", 1, 4, "what is up"]
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)

        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [data], "Body ok")

    async def test__insert_by_name(self):
        data = [1, "hello", 1, 4, "what is up"]
        res = await self.conn.insert(self.TESTER_SPACE_NAME, data)

        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [data], "Body ok")

    async def test__insert_by_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        data = [1, "hello", 1, 4, "what is up"]
        with self.assertRaises(TarantoolSchemaError):
            await self.conn.insert(self.TESTER_SPACE_NAME, data)

    async def test__insert_complex_tuple(self):
        p, p_cmp = get_complex_param(replace_bin=False)
        data = [1, "hello", 1, 2, p]
        data_cmp = [1, "hello", 1, 2, p_cmp]

        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertResponseEqual(res, [data_cmp], "Body ok")

    async def test__insert_replace(self):
        data = [1, "hello", 1, 4, "what is up"]

        await self.conn.insert(self.TESTER_SPACE_ID, data)

        try:
            data = [1, "hello2", 1, 4, "what is up"]
            res = await self.conn.insert(self.TESTER_SPACE_ID, t=data, replace=True)

            self.assertResponseEqual(res, [data], "Body ok")
        except Exception as e:
            self.fail(e)

    async def test__insert_invalid_types(self):
        with self.assertRaisesRegex(
            TypeError,
            r"missing 2 required positional arguments: " r"\'space\' and \'t\'",
        ):
            await self.conn.insert()

        with self.assertRaisesRegex(
            TypeError, r"missing 1 required positional argument: \'t\'"
        ):
            await self.conn.insert(self.TESTER_SPACE_ID)

    async def test__replace(self):
        data = [1, "hello", 1, 4, "what is up"]
        res = await self.conn.replace(self.TESTER_SPACE_ID, data)
        self.assertResponseEqual(res, [data], "Body ok")

        data = [1, "hello2", 1, 5, "what is up"]
        res = await self.conn.replace(self.TESTER_SPACE_ID, data)
        self.assertResponseEqual(res, [data], "Body ok")

    async def test__replace_invalid_types(self):
        with self.assertRaisesRegex(
            TypeError,
            r"missing 2 required positional arguments: " r"\'space\' and \'t\'",
        ):
            await self.conn.replace()

        with self.assertRaisesRegex(
            TypeError, r"missing 1 required positional argument: \'t\'"
        ):
            await self.conn.replace(self.TESTER_SPACE_ID)

    async def test__insert_dict_key(self):
        data = {
            "f1": 1,
            "f2": "hello",
            "f3": 5,
            "f4": 6,
            "f5": "hello dog",
        }
        data_cmp = [1, "hello", 5, 6, "hello dog"]
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertResponseEqual(res, [data_cmp], "Body ok")

    async def test__insert_dict_key_holes(self):
        data = {
            "f1": 1,
            "f2": "hello",
            "f3": 3,
            "f4": 6,
            "f5": None,
        }
        data_cmp = [1, "hello", 3, 6, None]
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertResponseEqual(res, [data_cmp], "Body ok")

    async def test__insert_no_special_empty_key(self):
        data = {
            "f1": 1,
            "f2": "hello",
            "f3": 3,
            "f4": 6,
            "f5": None,
        }
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)

        with self.assertRaises(KeyError):
            res[0][""]

    async def test__insert_dict_resp(self):
        data = [0, "hello", 0, 5, "wow"]
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertResponseEqualKV(
            res, [{"f1": 0, "f2": "hello", "f3": 0, "f4": 5, "f5": "wow"}]
        )

    async def test__insert_resp_extra(self):
        data = [0, "hello", 5, 6, "help", "common", "yo"]
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertResponseEqual(res, [data])

    async def test__insert_bin_as_str(self):
        try:
            (await self.conn.call("func_load_bin_str"))[0]
        except UnicodeDecodeError as e:
            self.fail(e)



import asyncio

from asynctnt.exceptions import ErrorCode, TarantoolDatabaseError
from tests import BaseTarantoolTestCase
from tests._testbase import ensure_bin_version, ensure_version


@ensure_bin_version(min=(2, 10))
class StreamTestCase(BaseTarantoolTestCase):
    EXTRA_BOX_CFG = "memtx_use_mvcc_engine = true"

    @ensure_version(min=(2, 10))
    async def test__transaction_commit(self):
        s = self.conn.stream()
        self.assertGreater(s.stream_id, 0)

        await s.begin()
        data = [1, "hello", 1, 4, "what is up"]
        await s.insert(self.TESTER_SPACE_NAME, data)
        res = await s.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

        await s.commit()

        res = await self.conn.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

    @ensure_version(min=(2, 10))
    async def test__transaction_rolled_back(self):
        s = self.conn.stream()
        await s.begin()
        data = [1, "hello", 1, 4, "what is up"]
        await s.insert(self.TESTER_SPACE_NAME, data)
        res = await s.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

        await s.rollback()

        res = await self.conn.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [])

    @ensure_version(min=(2, 10))
    async def test__transaction_begin_through_call(self):
        s = self.conn.stream()
        await s.call("box.begin")
        data = [1, "hello", 1, 4, "what is up"]
        await s.insert(self.TESTER_SPACE_NAME, data)
        res = await s.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

        await s.commit()

        res = await self.conn.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

    @ensure_version(min=(2, 10))
    async def test__transaction_commit_through_call(self):
        s = self.conn.stream()
        await s.begin()
        data = [1, "hello", 1, 4, "what is up"]
        await s.insert(self.TESTER_SPACE_NAME, data)
        res = await s.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

        await s.call("box.commit")

        res = await self.conn.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

    @ensure_version(min=(2, 10))
    async def test__transaction_rolled_back_through_call(self):
        s = self.conn.stream()
        await s.begin()
        data = [1, "hello", 1, 4, "what is up"]
        await s.insert(self.TESTER_SPACE_NAME, data)
        res = await s.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

        await s.call("box.rollback")

        res = await self.conn.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [])

    @ensure_version(min=(2, 10))
    async def test__transaction_commit_through_sql(self):
        s = self.conn.stream()
        await s.execute("START TRANSACTION")
        data = [1, "hello", 1, 4, "what is up"]
        await s.insert(self.TESTER_SPACE_NAME, data)
        res = await s.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

        await s.execute("COMMIT")

        res = await self.conn.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

    @ensure_version(min=(2, 10))
    async def test__transaction_rollback_through_sql(self):
        s = self.conn.stream()
        await s.execute("START TRANSACTION")
        data = [1, "hello", 1, 4, "what is up"]
        await s.insert(self.TESTER_SPACE_NAME, data)
        res = await s.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

        await s.execute("ROLLBACK")

        res = await self.conn.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [])

    @ensure_version(min=(2, 10))
    async def test__transaction_context_manager_commit(self):
        data = [1, "hello", 1, 4, "what is up"]

        async with self.conn.stream() as s:
            await s.insert(self.TESTER_SPACE_NAME, data)
            res = await s.select(self.TESTER_SPACE_NAME)
            self.assertResponseEqual(res, [data])

        res = await self.conn.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data])

    @ensure_version(min=(2, 10))
    async def test__transaction_context_manager_rollback(self):
        class ExpectedError(Exception):
            pass

        data = [1, "hello", 1, 4, "what is up"]

        try:
            async with self.conn.stream() as s:
                await s.insert(self.TESTER_SPACE_NAME, data)
                res = await s.select(self.TESTER_SPACE_NAME)
                self.assertResponseEqual(res, [data])

                raise ExpectedError("some error")
        except ExpectedError:
            pass

        res = await self.conn.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [])

    @ensure_version(min=(2, 10))
    async def test__transaction_2_streams(self):
        data1 = [1, "hello", 1, 4, "what is up"]
        data2 = [2, "hi", 100, 400, "nothing match"]

        s1 = self.conn.stream()
        s2 = self.conn.stream()

        await s1.begin()
        await s2.begin()

        await s1.insert(self.TESTER_SPACE_NAME, data1)
        await s2.insert(self.TESTER_SPACE_NAME, data2)

        res = await s1.select(self.TESTER_SPACE_NAME, [1])
        self.assertResponseEqual(res, [data1])

        res = await s2.select(self.TESTER_SPACE_NAME, [2])
        self.assertResponseEqual(res, [data2])

        await s1.commit()
        await s2.commit()

        res = await self.conn.select(self.TESTER_SPACE_NAME)
        self.assertResponseEqual(res, [data1, data2])

    @ensure_version(min=(2, 10))
    async def test__transaction_timeout(self):
        s = self.conn.stream()
        await s.begin(tx_timeout=0.5)

        await asyncio.sleep(1.0)

        with self.assertRaises(TarantoolDatabaseError) as exc:
            await s.commit()

        self.assertEqual(ErrorCode.ER_TRANSACTION_TIMEOUT, exc.exception.code)
        self.assertEqual(
            "Transaction has been aborted by timeout", exc.exception.message
        )



cimport cython


@cython.final
cdef class UpdateRequest(BaseRequest):
    cdef int encode_body(self, WriteBuffer buffer) except -1:
        return encode_request_update(buffer, self.space, self.index,
                                     self.key, self.operations, <bint> False)

cdef char *encode_update_ops(WriteBuffer buffer,
                             char *p, list operations,
                             SchemaSpace space) except NULL:
    cdef:
        char *begin
        uint32_t ops_len, op_len
        bytes str_temp
        char *str_c
        ssize_t str_len

        char *op_str_c
        ssize_t op_str_len
        char op

        uint32_t extra_length

        bint field_encode_as_str
        uint64_t field_no
        char *field_str_c
        ssize_t field_str_len
        object field_no_obj

        uint32_t splice_position, splice_offset

    field_encode_as_str = 0
    field_str_c = NULL

    begin = NULL

    if operations is not None:
        ops_len = <uint32_t> cpython.list.PyList_GET_SIZE(operations)
    else:
        ops_len = 0

    p = buffer.mp_encode_array(p, ops_len)
    if ops_len == 0:
        return p

    for operation in operations:
        if isinstance(operation, tuple):
            op_len = cpython.tuple.PyTuple_GET_SIZE(operation)
        elif isinstance(operation, list):
            op_len = cpython.list.PyList_GET_SIZE(operation)
        else:
            raise TypeError(
                'Single operation must be a tuple or list')
        if op_len < 3:
            raise IndexError(
                'Operation length must be at least 3')

        op_type_str = operation[0]
        if isinstance(op_type_str, str):
            str_temp = encode_unicode_string(op_type_str, buffer._encoding)
        elif isinstance(op_type_str, bytes):
            str_temp = <bytes> op_type_str
        else:
            raise TypeError(
                'Operation type must of a str or bytes type')

        cpython.bytes.PyBytes_AsStringAndSize(str_temp, &op_str_c,
                                              &op_str_len)

        field_no_obj = operation[1]
        if isinstance(field_no_obj, int):
            field_no = <int> field_no_obj
        elif isinstance(field_no_obj, str):
            if space.metadata is not None:
                field_no = <int> space.metadata.id_by_name_safe(field_no_obj)
                if field_no == -1:
                    field_encode_as_str = 1
            else:
                field_encode_as_str = 1

            if field_encode_as_str:
                str_temp = encode_unicode_string(field_no_obj, buffer._encoding)
                cpython.bytes.PyBytes_AsStringAndSize(str_temp, &field_str_c, &field_str_len)
        else:
            raise TypeError(
                'Operation field_no must be of either int or str type')

        op = <char> 0
        if op_str_len == 1:
            op = op_str_c[0]

        if op == tarantool.IPROTO_OP_ADD \
                or op == tarantool.IPROTO_OP_SUB \
                or op == tarantool.IPROTO_OP_AND \
                or op == tarantool.IPROTO_OP_XOR \
                or op == tarantool.IPROTO_OP_OR \
                or op == tarantool.IPROTO_OP_DELETE \
                or op == tarantool.IPROTO_OP_INSERT \
                or op == tarantool.IPROTO_OP_ASSIGN:
            # mp_sizeof_array(3)
            # + mp_sizeof_str(1)
            # + mp_sizeof_uint(field_no)
            extra_length = 1 + 2 + mp_sizeof_uint(field_no)
            p = begin = buffer._ensure_allocated(p, extra_length)

            p = mp_encode_array(p, 3)
            p = mp_encode_str(p, op_str_c, 1)
            if field_str_c == NULL:
                p = mp_encode_uint(p, field_no)
            else:
                p = mp_encode_str(p, field_str_c, field_str_len)

            buffer._length += (p - begin)

            op_argument = operation[2]
            p = buffer.mp_encode_obj(p, op_argument)

        elif op == tarantool.IPROTO_OP_SPLICE:
            if op_len < 5:
                raise ValueError(
                    'Splice operation must have length of 5, '
                    'but got: {}'.format(op_len)
                )

            splice_position_obj = operation[2]
            splice_offset_obj = operation[3]
            op_argument = operation[4]
            if not isinstance(splice_position_obj, int):
                raise TypeError('Splice position must be int')
            if not isinstance(splice_offset_obj, int):
                raise TypeError('Splice offset must be int')

            splice_position = <uint32_t> splice_position_obj
            splice_offset = <uint32_t> splice_offset_obj

            # mp_sizeof_array(5) + mp_sizeof_str(1) + ...
            extra_length = 1 + 2 \
                           + mp_sizeof_uint(field_no) \
                           + mp_sizeof_uint(splice_position) \
                           + mp_sizeof_uint(splice_offset)
            p = begin = buffer._ensure_allocated(p, extra_length)

            p = mp_encode_array(p, 5)
            p = mp_encode_str(p, op_str_c, 1)
            if field_str_c == NULL:
                p = mp_encode_uint(p, field_no)
            else:
                p = mp_encode_str(p, field_str_c, field_str_len)
            p = mp_encode_uint(p, splice_position)
            p = mp_encode_uint(p, splice_offset)
            buffer._length += (p - begin)
            p = buffer.mp_encode_obj(p, op_argument)
        else:
            raise TypeError(
                'Unknown update operation type `{}`'.format(op_type_str))
    return p

cdef int encode_request_update(WriteBuffer buffer,
                               SchemaSpace space, SchemaIndex index,
                               key_tuple, list operations,
                               bint is_upsert) except -1:
    cdef:
        char *begin
        char *p
        uint32_t body_map_sz
        uint32_t max_body_len
        uint32_t space_id, index_id
        uint32_t key_of_tuple, key_of_operations
        Metadata metadata
        bint default_fields_none

    space_id = space.sid
    index_id = index.iid

    if not is_upsert:
        key_of_tuple = tarantool.IPROTO_KEY
        key_of_operations = tarantool.IPROTO_TUPLE
        metadata = index.metadata
        default_fields_none = False
    else:
        key_of_tuple = tarantool.IPROTO_TUPLE
        key_of_operations = tarantool.IPROTO_OPS
        metadata = space.metadata
        default_fields_none = True

    body_map_sz = 3 + <uint32_t> (index_id > 0)
    # Size description:
    # mp_sizeof_map(body_map_sz)
    # + mp_sizeof_uint(TP_SPACE)
    # + mp_sizeof_uint(space)
    max_body_len = 1 \
                   + 1 \
                   + 9

    if index_id > 0:
        # + mp_sizeof_uint(TP_INDEX)
        # + mp_sizeof_uint(index)
        max_body_len += 1 + 9

    max_body_len += 1  # + mp_sizeof_uint(TP_KEY)
    max_body_len += 1  # + mp_sizeof_uint(TP_TUPLE)

    buffer.ensure_allocated(max_body_len)

    p = begin = &buffer._buf[buffer._length]
    p = mp_encode_map(p, body_map_sz)
    p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
    p = mp_encode_uint(p, space_id)

    if index_id > 0:
        p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
        p = mp_encode_uint(p, index_id)
    buffer._length += (p - begin)

    p = buffer.mp_encode_uint(p, key_of_tuple)
    p = encode_key_sequence(buffer, p, key_tuple, metadata, default_fields_none)

    p = buffer.mp_encode_uint(p, key_of_operations)
    p = encode_update_ops(buffer, p, operations, space)



DEF _BUFFER_FREELIST_SIZE = 256
DEF _BUFFER_INITIAL_SIZE = 1024
DEF _BUFFER_MAX_GROW = 65536

DEF _DEALLOCATE_RATIO = 4

DEF METADATA_FREELIST_SIZE = 128
DEF REQUEST_FREELIST = 256

# Header length description:
# pkt_len +
# mp_sizeof_map(2) +
# mp_sizeof_uint(TP_CODE) +
# mp_sizeof_uint(TP COMMAND) +
# mp_sizeof_uint(TP_SYNC) +
# sync len +
# mp_sizeof_uint(TP_SCHEMA_ID) +
# mp_sizeof_uint(STREAM_ID) +
# schema_id len
DEF HEADER_CONST_LEN = 5 + 1 + 1 + 1 + 1 + 5 + 1 + 9 + 9

DEF IPROTO_GREETING_SIZE = 128
DEF TARANTOOL_VERSION_LENGTH = 64
DEF SALT_LENGTH = 44
DEF SCRAMBLE_SIZE = 20


DEF SPACE_VSPACE = 281
DEF SPACE_VINDEX = 289

DEF DATETIME_TAIL_SZ = 4 + 2 + 2



cdef class BeginRequest(BaseRequest):
    cdef:
        uint32_t isolation
        double tx_timeout

cdef class CommitRequest(BaseRequest):
    pass

cdef class RollbackRequest(BaseRequest):
    pass



cimport cython
from libc.stdint cimport uint32_t, uint64_t


@cython.final
cdef class Db:
    cdef:
        uint64_t _stream_id
        BaseProtocol _protocol
        bytes _encoding

    @staticmethod
    cdef inline Db create(BaseProtocol protocol, uint64_t stream_id)

    cdef inline uint64_t next_sync(self)

    cdef object _ping(self, float timeout)

    cdef object _id(self, float timeout)

    cdef object _auth(self,
                      bytes salt,
                      str username,
                      str password,
                      float timeout)

    cdef object _call(self,
                      tarantool.iproto_type op,
                      str func_name,
                      object args,
                      float timeout,
                      bint push_subscribe)

    cdef object _eval(self,
                      str expression,
                      object args,
                      float timeout,
                      bint push_subscribe)

    cdef object _select(self,
                        object space,
                        object index,
                        object key,
                        uint64_t offset,
                        uint64_t limit,
                        object iterator,
                        float timeout,
                        bint check_schema_change)

    cdef object _insert(self,
                        object space,
                        object t,
                        bint replace,
                        float timeout)

    cdef object _delete(self,
                        object space,
                        object index,
                        object key,
                        float timeout)

    cdef object _update(self,
                        object space,
                        object index,
                        object key,
                        list operations,
                        float timeout)

    cdef object _upsert(self,
                        object space,
                        object t,
                        list operations,
                        float timeout)

    cdef object _execute(self,
                         query,
                         object args,
                         bint parse_metadata,
                         float timeout)

    cdef object _prepare(self,
                         query,
                         bint parse_metadata,
                         float timeout)

    cdef object _begin(self,
                       uint32_t isolation,
                       double tx_timeout,
                       float timeout)

    cdef object _commit(self, float timeout)

    cdef object _rollback(self, float timeout)



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from .iproto import protocol

if TYPE_CHECKING:  # pragma: nocover
    from .api import Api


class PreparedStatement:
    __slots__ = ("_api", "_query", "_stmt_id", "_params", "_params_count")

    def __init__(self, api: "Api", query: str):
        self._api = api
        self._query = query
        self._stmt_id = None
        self._params = None
        self._params_count = 0

    @property
    def id(self) -> int:
        """
        Prepared statement id
        """
        return self._stmt_id

    @property
    def params_count(self) -> int:
        """
        Bound params count
        """
        return self._params_count

    @property
    def params(self) -> Optional[protocol.Metadata]:
        """
        Bound params metadata
        """
        return self._params

    async def prepare(self, timeout: float = -1.0) -> int:
        """
            Prepare statement

        :param timeout: request timeout
        :return: prepared statement id
        """
        resp = await self._api.prepare_iproto(self._query, timeout=timeout)
        self._stmt_id = resp.stmt_id
        self._params = resp.params
        self._params_count = resp.params_count
        return self._stmt_id

    async def execute(
        self,
        args: Optional[List[Union[Dict[str, Any], Any]]] = None,
        *,
        parse_metadata: bool = True,
        timeout: float = -1.0,
    ) -> protocol.Response:
        """
            Execute this prepared statement with specified args
        :param args: arguments list
        :param parse_metadata: whether to parse response metadata or not
        :param timeout: request timeout
        """
        return await self._api.execute(
            query=self._stmt_id,
            args=args,
            parse_metadata=parse_metadata,
            timeout=timeout,
        )

    async def unprepare(self, timeout: float = -1.0):
        """
            Unprepare current prepared statement
        :param timeout: request timeout
        """
        await self._api.unprepare_iproto(self._stmt_id, timeout=timeout)
        self._stmt_id = None

    async def __aenter__(self):
        """
        If used as a Context Manager `prepare()` and `unprepare()` methods
        are called automatically
        """
        if self._stmt_id is None:
            await self.prepare()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._stmt_id is not None:
            await self.unprepare()



cimport cpython
cimport cpython.bytearray
cimport cpython.bytes
cimport cpython.dict
from cpython.ref cimport PyObject
from libc.stdint cimport uint32_t, uint64_t

import base64
import re
import socket

from asynctnt.exceptions import TarantoolDatabaseError
from asynctnt.log import logger

VERSION_STRING_REGEX = re.compile(r'\s*Tarantool\s+([\d.]+)\s+.*')


cdef class CoreProtocol:
    def __init__(self,
                 host, port,
                 encoding=None,
                 initial_read_buffer_size=None):
        self.host = host
        self.port = port

        encoding = encoding or b'utf-8'
        if isinstance(encoding, str):
            self.encoding = encoding.encode()
        elif isinstance(self.encoding, bytes):
            self.encoding = encoding
        else:
            raise TypeError('encoding must be either str or bytes')

        initial_read_buffer_size = initial_read_buffer_size or 0x20000
        self.transport = None

        self.rbuf = ReadBuffer.create(encoding, initial_read_buffer_size)
        self.state = PROTOCOL_IDLE
        self.con_state = CONNECTION_BAD

        self.version = None
        self.salt = None

    cdef bint _is_connected(self):
        return self.con_state != CONNECTION_BAD

    cdef bint _is_fully_connected(self):
        return self.con_state == CONNECTION_FULL

    def is_connected(self):
        return self._is_connected()

    def is_fully_connected(self):
        return self._is_fully_connected()

    def get_version(self):
        return self.version

    cdef void _write(self, buf) except *:
        self.transport.write(memoryview(buf))

    cdef void _on_data_received(self, data):
        cdef:
            size_t ruse, curr
            const char *p
            const char *q
            uint32_t packet_len

            char *data_str
            ssize_t data_len
            ssize_t buf_len

        data_str = NULL
        data_len = 0

        if cpython.PyBytes_CheckExact(data):
            cpython.bytes.PyBytes_AsStringAndSize(<bytes> data,
                                                  &data_str,
                                                  &data_len)
            self.rbuf.extend(data_str, data_len)
        elif cpython.bytearray.PyByteArray_CheckExact(data):
            data_len = cpython.bytearray.PyByteArray_Size(<bytearray> data)
            self.rbuf.extend(cpython.bytearray.PyByteArray_AsString(<bytearray> data), data_len)
        else:
            raise BufferError(
                '_on_data_received: expected bytes or bytearray object, got {}'.format(
                    type(data),
                ))

        if data_len == 0:
            return

        if self.state == PROTOCOL_GREETING:
            if self.rbuf.use < IPROTO_GREETING_SIZE:
                # not enough for greeting
                return
            self._process__greeting()
            self.rbuf.move(IPROTO_GREETING_SIZE)
        elif self.state == PROTOCOL_NORMAL:
            p = self.rbuf.buf
            end = &self.rbuf.buf[self.rbuf.use]

            while p < end:
                q = p  # q is temporary to parse packet length
                buf_len = end - p
                if buf_len < 5:
                    # not enough
                    break

                q = &q[1]  # skip to 2nd byte of packet length
                packet_len = mp_load_u32(&q)

                if buf_len < 5 + packet_len:
                    # not enough to read an entire packet
                    break

                p = &p[5]  # skip length header
                self._on_response_received(p, packet_len)
                p = &p[packet_len]

                if p == end:
                    self.rbuf.use = 0
                    break

            self.rbuf.use = end - p
            if self.rbuf.use > 0:
                self.rbuf.move_offset(p - self.rbuf.buf, self.rbuf.use)
        else:
            # TODO: raise exception
            pass

    cdef void _process__greeting(self):
        cdef size_t ver_length = TARANTOOL_VERSION_LENGTH
        rbuf = self.rbuf
        self.version = self._parse_version(self.rbuf.get_slice_end(ver_length))
        self.salt = base64.b64decode(
            self.rbuf.get_slice(ver_length,
                                ver_length + SALT_LENGTH)
        )[:SCRAMBLE_SIZE]
        self.state = PROTOCOL_NORMAL
        self._on_greeting_received()

    def _parse_version(self, version):
        m = VERSION_STRING_REGEX.match(version.decode('ascii'))
        if m is not None:
            ver = m.group(1)
            return tuple(map(int, ver.split('.')))

    cdef void _on_greeting_received(self):
        pass

    cdef void _on_response_received(self, const char *buf, uint32_t buf_len):
        pass

    cdef void _on_connection_made(self):
        pass

    cdef void _on_connection_lost(self, exc):
        pass

    # asyncio callbacks

    def data_received(self, data):
        self._on_data_received(data)

    def connection_made(self, transport):
        self.transport = transport
        self.con_state = CONNECTION_CONNECTED

        sock = transport.get_extra_info('socket')
        if sock is not None \
                and (not hasattr(socket, 'AF_UNIX')
                     or sock.family != socket.AF_UNIX):
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        self.state = PROTOCOL_GREETING
        self._on_connection_made()

    def connection_lost(self, exc):
        self.con_state = CONNECTION_BAD
        self.version = None
        self.salt = None
        self.rbuf = None

        self._on_connection_lost(exc)



cimport cython


@cython.final
cdef class InsertRequest(BaseRequest):

    cdef int encode_body(self, WriteBuffer buffer) except -1:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len
            uint32_t space_id

        space_id = self.space.sid

        body_map_sz = 2
        # Size description:
        # mp_sizeof_map(body_map_sz)
        # + mp_sizeof_uint(TP_SPACE)
        # + mp_sizeof_uint(space)
        # + mp_sizeof_uint(TP_TUPLE)
        max_body_len = 1 \
                       + 1 \
                       + 9 \
                       + 1

        buffer.ensure_allocated(max_body_len)

        p = begin = &buffer._buf[buffer._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
        p = mp_encode_uint(p, space_id)

        p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
        buffer._length += (p - begin)
        p = encode_key_sequence(buffer, p, self.t, self.space.metadata, True)



cimport cython
from libc.stdint cimport int64_t, uint64_t


@cython.freelist(REQUEST_FREELIST)
cdef class BaseRequest:

    # def __cinit__(self):
    #     self.sync = 0
    #     self.schema_id = -1
    #     self.parse_as_tuples = False
    #     self.parse_metadata = True
    #     self.push_subscribe = False

    cdef inline WriteBuffer encode(self, bytes encoding):
        cdef WriteBuffer buffer = WriteBuffer.create(encoding)
        buffer.write_header(self.sync, self.op, self.schema_id, self.stream_id)
        self.encode_body(buffer)
        buffer.write_length()
        return buffer

    cdef int encode_body(self, WriteBuffer buffer) except -1:
        return 0

    def __repr__(self):  # pragma: nocover
        return \
            '<Request op={} sync={} schema_id={} push_subscribe={}>'.format(
                self.op,
                self.sync,
                self.schema_id,
                self.push_subscribe
            )


cdef char *encode_key_sequence(WriteBuffer buffer,
                               char *p, object t,
                               Metadata metadata,
                               bint default_none) except NULL:
    if isinstance(t, list) or t is None:
        return buffer.mp_encode_list(p, <list> t)
    elif isinstance(t, tuple):
        return buffer.mp_encode_tuple(p, <tuple> t)
    elif isinstance(t, dict) and metadata is not None:
        return buffer.mp_encode_list(
            p, dict_to_list_fields(<dict> t, metadata, default_none)
        )
    else:
        if metadata is not None:
            msg = 'sequence must be either list, tuple or dict'
        else:
            msg = 'sequence must be either list or tuple'
        raise TypeError(
            '{}, got: {}'.format(msg, type(t))
        )



import asyncio
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

class Field:
    name: Optional[str]
    """ Field name """

    type: Optional[str]
    """ Field type """

    collation: Optional[str]
    """ Field collation value """

    is_nullable: Optional[bool]
    """ If field may be null """

    is_autoincrement: Optional[bool]
    """ Is Autoincrement """

    span: Optional[str]

class Metadata:
    fields: List[Field]
    """ List of fields """

    name_id_map: Dict[str, int]
    """ Mapping name -> id """

class SchemaIndex:
    iid: int
    """ Index id """

    sid: int
    """ Space id """

    name: Optional[str]
    index_type: Optional[str]
    unique: Optional[bool]
    metadata: Optional[Metadata]

class SchemaSpace:
    sid: int
    owner: int
    name: Optional[str]
    engine: Optional[str]
    field_count: int
    flags: Optional[Any]
    metadata: Optional[Metadata]
    indexes: Dict[Union[int, str], SchemaIndex]

class Schema:
    id: int
    spaces: Dict[Union[str, int], SchemaSpace]

class TarantoolTuple:
    def __repr__(self) -> str: ...
    def __index__(self, i: int) -> Any: ...
    def __len__(self) -> int: ...
    def __contains__(self, item: str) -> bool: ...
    def __getitem__(self, item: Union[int, str, slice]) -> Any: ...
    def keys(self) -> Iterator[str]: ...
    def values(self) -> Iterator[Any]: ...
    def items(self) -> Iterator[Tuple[str, Any]]: ...
    def get(self, item: str) -> Optional[Any]: ...
    def __iter__(self): ...
    def __next__(self): ...

class IProtoErrorStackFrame:
    error_type: str
    file: str
    line: int
    message: str
    err_no: int
    code: int
    fields: Dict[str, Any]

class IProtoError:
    trace: List[IProtoErrorStackFrame]

BodyItem = Union[TarantoolTuple, List[Any], Dict[Any, Any], Any]

class Response:
    errmsg: Optional[str]
    error: Optional[IProtoError]
    encoding: bytes
    autoincrement_ids: Optional[List[int]]
    body: Optional[List[BodyItem]]
    metadata: Optional[Metadata]
    params: Optional[Metadata]
    params_count: int

    @property
    def sync(self) -> int: ...
    @property
    def code(self) -> int: ...
    @property
    def return_code(self) -> int: ...
    @property
    def schema_id(self) -> int: ...
    @property
    def stmt_id(self) -> int: ...
    @property
    def rowcount(self) -> int: ...
    def done(self) -> bool: ...
    def __len__(self) -> int: ...
    def __getitem__(self, i) -> BodyItem: ...
    def __iter__(self): ...

class PushIterator:
    def __init__(self, fut: asyncio.Future): ...
    def __iter__(self): ...
    def __next__(self): ...
    def __aiter__(self): ...
    async def __anext__(self): ...
    @property
    def response(self) -> Response: ...

class Db:
    @property
    def stream_id(self) -> int: ...
    def set_stream_id(self, stream_id: int): ...
    def ping(self, timeout: float = -1): ...
    def call16(
        self,
        func_name: str,
        args=None,
        timeout: float = -1,
        push_subscribe: bool = False,
    ): ...
    def call(
        self,
        func_name: str,
        args=None,
        timeout: float = -1,
        push_subscribe: bool = False,
    ): ...
    def eval(
        self,
        expression: str,
        args=None,
        timeout: float = -1,
        push_subscribe: bool = False,
    ): ...
    def select(
        self,
        space,
        key=None,
        offset: int = 0,
        limit: int = 0xFFFFFFFF,
        index=0,
        iterator=0,
        timeout: float = -1,
        check_schema_change: bool = True,
    ): ...
    def insert(self, space, t, replace: bool = False, timeout: float = -1): ...
    def replace(self, space, t, timeout: float = -1): ...
    def delete(self, space, key, index=0, timeout: float = -1): ...
    def update(self, space, key, operations, index=0, timeout: float = -1): ...
    def upsert(self, space, t, operations, timeout: float = -1): ...
    def execute(
        self, query, args, parse_metadata: bool = True, timeout: float = -1
    ): ...
    def prepare(self, query, parse_metadata: bool = True, timeout: float = -1): ...
    def begin(self, isolation: int, tx_timeout: float, timeout: float = -1): ...
    def commit(self, timeout: float = -1): ...
    def rollback(self, timeout: float = -1): ...

class Protocol:
    @property
    def schema_id(self) -> int: ...
    @property
    def schema(self) -> Schema: ...
    def create_db(self, gen_stream_id: bool = False) -> Db: ...
    def get_common_db(self) -> Db: ...
    def refetch_schema(self) -> asyncio.Future: ...
    def is_connected(self) -> bool: ...
    def is_fully_connected(self) -> bool: ...
    def get_version(self) -> tuple: ...



cdef enum iproto_header_key:
    IPROTO_REQUEST_TYPE = 0x00
    IPROTO_SYNC = 0x01
    IPROTO_REPLICA_ID = 0x02
    IPROTO_LSN = 0x03
    IPROTO_TIMESTAMP = 0x04
    IPROTO_SCHEMA_VERSION = 0x05
    IPROTO_SERVER_VERSION = 0x06
    IPROTO_GROUP_ID = 0x07
    IPROTO_STREAM_ID=0x0a


cdef enum iproto_key:
    IPROTO_SPACE_ID = 0x10
    IPROTO_INDEX_ID = 0x11
    IPROTO_LIMIT = 0x12
    IPROTO_OFFSET = 0x13
    IPROTO_ITERATOR = 0x14
    IPROTO_INDEX_BASE = 0x15

    IPROTO_KEY = 0x20
    IPROTO_TUPLE = 0x21
    IPROTO_FUNCTION_NAME = 0x22
    IPROTO_USER_NAME = 0x23
    IPROTO_INSTANCE_UUID = 0x24
    IPROTO_CLUSTER_UUID = 0x25
    IPROTO_VCLOCK = 0x26
    IPROTO_EXPR = 0x27
    IPROTO_OPS = 0x28
    IPROTO_BALLOT = 0x29
    IPROTO_TUPLE_META = 0x2a
    IPROTO_OPTIONS = 0x2b

    IPROTO_DATA = 0x30
    IPROTO_ERROR_24 = 0x31
    IPROTO_METADATA = 0x32
    IPROTO_BIND_METADATA = 0x33
    IPROTO_BIND_COUNT = 0x34

    IPROTO_SQL_TEXT = 0x40
    IPROTO_SQL_BIND = 0x41
    IPROTO_SQL_INFO = 0x42
    IPROTO_STMT_ID = 0x43

    IPROTO_ERROR = 0x52
    IPROTO_VERSION = 0x54
    IPROTO_FEATURES = 0x55
    IPROTO_AUTH_TYPE = 0x5b
    IPROTO_TIMEOUT = 0x56
    IPROTO_TXN_ISOLATION = 0x59

    IPROTO_CHUNK = 0x80


cdef enum iproto_metadata_key:
    IPROTO_FIELD_NAME = 0x00
    IPROTO_FIELD_TYPE = 0x01
    IPROTO_FIELD_COLL = 0x02
    IPROTO_FIELD_IS_NULLABLE = 0x03
    IPROTO_FIELD_IS_AUTOINCREMENT = 0x04
    IPROTO_FIELD_SPAN = 0x05


cdef enum iproto_sql_info_key:
    SQL_INFO_ROW_COUNT = 0x00
    SQL_INFO_AUTOINCREMENT_IDS = 0x01


cdef enum iproto_type:
    IPROTO_SELECT = 0x01
    IPROTO_INSERT = 0x02
    IPROTO_REPLACE = 0x03
    IPROTO_UPDATE = 0x04
    IPROTO_DELETE = 0x05
    IPROTO_CALL_16 = 0x06
    IPROTO_AUTH = 0x07
    IPROTO_EVAL = 0x08
    IPROTO_UPSERT = 0x09
    IPROTO_CALL = 0x0a
    IPROTO_EXECUTE = 0x0b
    IPROTO_PREPARE = 0x0d
    IPROTO_BEGIN = 0x0e
    IPROTO_COMMIT = 0x0f
    IPROTO_ROLLBACK = 0x10
    IPROTO_PING = 0x40
    IPROTO_ID = 0x49


cdef enum iproto_update_operation:
    IPROTO_OP_ADD = 43  # ord('+')
    IPROTO_OP_SUB = 45  # ord('-')
    IPROTO_OP_AND = 38  # ord('&')
    IPROTO_OP_XOR = 94  # ord('^')
    IPROTO_OP_OR = 124  # ord('|')
    IPROTO_OP_DELETE = 35  # ord('#')
    IPROTO_OP_INSERT = 33  # ord('!')
    IPROTO_OP_ASSIGN = 61  # ord('=')
    IPROTO_OP_SPLICE = 58  # ord(':')


cdef enum mp_extension_type:
    MP_UNKNOWN_EXTENSION = 0
    MP_DECIMAL = 1
    MP_UUID = 2
    MP_ERROR = 3
    MP_DATETIME = 4

cdef enum iproto_features:
    IPROTO_FEATURE_STREAMS = 0
    IPROTO_FEATURE_TRANSACTIONS = 1
    IPROTO_FEATURE_ERROR_EXTENSION = 2
    IPROTO_FEATURE_WATCHERS = 3

cdef enum iproto_error_fields:
    MP_ERROR_STACK = 0x00

cdef enum iproto_error_stack_fields:
    MP_ERROR_TYPE = 0x00
    MP_ERROR_FILE = 0x01
    MP_ERROR_LINE = 0x02
    MP_ERROR_MESSAGE = 0x03
    MP_ERROR_ERRNO = 0x04
    MP_ERROR_ERRCODE = 0x05
    MP_ERROR_FIELDS = 0x06



import copy


def get_complex_param(replace_bin=True, encoding="utf-8"):
    p = {
        "a": 1,
        "b": 2.5,
        "c": [1, 2, [4, 5], {"3": 17}],
        "d": {"k": 1, "l": (1, 2)},
        "e": b"1234567890",
        "f": [],
        "g": {},
        "h": "some long text",
        "i": "русский текст",
        "j": False,
        "k": True,
        "l": None,
    }
    p_copy = copy.deepcopy(p)
    # tuples return as lists
    p_copy["d"]["l"] = list(p_copy["d"]["l"])

    if replace_bin:
        # For some reason Tarantool in call returns MP_STR instead of MP_BIN
        p_copy["e"] = p_copy["e"].decode(encoding)

    return p, p_copy


def get_big_param(size=1024, parts=3):
    assert parts > 0
    d = {}
    for p in range(parts):
        d["a" + str(p)] = "x" * (size // parts)
    return d



# -*- coding: utf-8 -*-
import os
import re

import setuptools
from setuptools.command import build_ext as setuptools_build_ext


def find_version():
    for line in open("asynctnt/__init__.py"):
        if line.startswith("__version__"):
            return re.match(r"""__version__\s*=\s*(['"])([^'"]+)\1""", line).group(2)


CYTHON_VERSION = "3.0.7"


class build_ext(setuptools_build_ext.build_ext):
    user_options = setuptools_build_ext.build_ext.user_options + [
        ("cython-always", None, "run cythonize() even if .c files are present"),
        (
            "cython-annotate",
            None,
            "Produce a colorized HTML version of the Cython source.",
        ),
        ("cython-directives=", None, "Cython compiler directives"),
    ]

    def initialize_options(self):
        if getattr(self, "_initialized", False):
            return

        super(build_ext, self).initialize_options()

        if os.environ.get("ASYNCTNT_DEBUG"):
            self.cython_always = True
            self.cython_annotate = True
            self.cython_directives = {
                "linetrace": True,
            }
            self.define = "CYTHON_TRACE,CYTHON_TRACE_NOGIL"
            self.debug = True
            self.gdb_debug = True
        else:
            self.cython_always = False
            self.cython_annotate = None
            self.cython_directives = None
            self.gdb_debug = False

    def finalize_options(self):
        if getattr(self, "_initialized", False):
            return

        need_cythonize = self.cython_always

        if not need_cythonize:
            for extension in self.distribution.ext_modules:
                for i, sfile in enumerate(extension.sources):
                    if sfile.endswith(".pyx"):
                        prefix, ext = os.path.splitext(sfile)
                        cfile = prefix + ".c"

                        if os.path.exists(cfile) and not self.cython_always:
                            extension.sources[i] = cfile
                        else:
                            need_cythonize = True

        if need_cythonize:
            self.cythonize()

        super(build_ext, self).finalize_options()

    def cythonize(self):
        try:
            import Cython
        except ImportError as e:
            raise RuntimeError(
                "please install Cython to compile asynctnt from source"
            ) from e

        if Cython.__version__ != CYTHON_VERSION:
            raise RuntimeError(
                "asynctnt requires Cython version {}, got {}".format(
                    CYTHON_VERSION, Cython.__version__
                )
            )

        from Cython.Build import cythonize

        directives = {"language_level": "3"}
        if self.cython_directives:
            if isinstance(self.cython_directives, str):
                for directive in self.cython_directives.split(","):
                    k, _, v = directive.partition("=")
                    if v.lower() == "false":
                        v = False
                    if v.lower() == "true":
                        v = True

                    directives[k] = v
            elif isinstance(self.cython_directives, dict):
                directives.update(self.cython_directives)

        self.distribution.ext_modules[:] = cythonize(
            self.distribution.ext_modules,
            compiler_directives=directives,
            annotate=self.cython_annotate,
            gdb_debug=self.gdb_debug,
        )


setuptools.setup(
    version=find_version(),
    cmdclass={"build_ext": build_ext},
    ext_modules=[
        setuptools.extension.Extension(
            "asynctnt.iproto.protocol",
            sources=[
                "asynctnt/iproto/protocol.pyx",
                "third_party/msgpuck/msgpuck.c",
                "third_party/msgpuck/hints.c",
                "asynctnt/iproto/tupleobj/tupleobj.c",
            ],
            include_dirs=[
                "third_party",
                "asynctnt/iproto",
            ],
        )
    ],
)



import asynctnt
from asynctnt import Response
from tests import BaseTarantoolTestCase
from tests._testbase import ensure_version


class SQLExecuteTestCase(BaseTarantoolTestCase):
    def _compat_field_name(self, field_name: str) -> str:
        if self.conn.version >= (3, 0):
            return field_name
        return field_name.upper()

    @ensure_version(min=(2, 0))
    async def test__sql_basic(self):
        res = await self.conn.execute("select 1, 2")

        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [[1, 2]], "Body ok")

    @ensure_version(min=(2, 0))
    async def test__sql_with_param(self):
        res = await self.conn.execute("select 1, 2 where 1 = ?", [1])

        self.assertResponseEqual(res, [[1, 2]], "Body ok")

    @ensure_version(min=(2, 0))
    async def test__sql_with_param_cols(self):
        res = await self.conn.execute("select 1 as a, 2 as b where 1 = ?", [1])

        self.assertResponseEqualKV(
            res,
            [
                {
                    self._compat_field_name("a"): 1,
                    self._compat_field_name("b"): 2,
                }
            ],
            "Body ok",
        )

    @ensure_version(min=(2, 0))
    async def test__sql_with_param_cols2(self):
        res = await self.conn.execute(
            "select 1 as a, 2 as b where 1 = ? and 2 = ?", [1, 2]
        )

        self.assertResponseEqualKV(
            res,
            [
                {
                    self._compat_field_name("a"): 1,
                    self._compat_field_name("b"): 2,
                }
            ],
            "Body ok",
        )

    @ensure_version(min=(2, 0))
    async def test__sql_with_param_cols_maps(self):
        res = await self.conn.execute(
            "select 1 as a, 2 as b where 1 = :p1 and 2 = :p2",
            [
                {":p1": 1},
                {":p2": 2},
            ],
        )

        self.assertResponseEqualKV(
            res,
            [
                {
                    self._compat_field_name("a"): 1,
                    self._compat_field_name("b"): 2,
                }
            ],
            "Body ok",
        )

    @ensure_version(min=(2, 0))
    async def test__sql_with_param_cols_maps_and_positional(self):
        res = await self.conn.execute(
            "select 1 as a, 2 as b where 1 = :p1 and 2 = :p2 and 3 = ? and 4 = ?",
            [{":p1": 1}, {":p2": 2}, 3, 4],
        )

        self.assertResponseEqualKV(
            res,
            [
                {
                    self._compat_field_name("a"): 1,
                    self._compat_field_name("b"): 2,
                }
            ],
            "Body ok",
        )

    @ensure_version(min=(2, 0))
    async def test__sql_insert(self):
        res = await self.conn.execute(
            "insert into sql_space (id, name) values (1, 'one')"
        )
        self.assertEqual(1, res.rowcount, "rowcount ok")

    @ensure_version(min=(2, 0))
    async def test__sql_empty_autoincrement(self):
        res = await self.conn.execute(
            "insert into sql_space (id, name) values (1, 'one')"
        )
        self.assertEqual(None, res.autoincrement_ids, "autoincrement ok")

    @ensure_version(min=(2, 0))
    async def test__sql_insert_autoincrement(self):
        res = await self.conn.execute(
            "insert into sql_space_autoincrement (name) values ('name')"
        )
        self.assertEqual(1, res.rowcount, "rowcount ok")
        self.assertEqual([1], res.autoincrement_ids, "autoincrement ok")

    @ensure_version(min=(2, 0))
    async def test__sql_insert_autoincrement_multiple(self):
        res = await self.conn.execute(
            "insert into sql_space_autoincrement_multiple (name) "
            "values ('name'), ('name2')"
        )
        self.assertEqual(2, res.rowcount, "rowcount ok")
        self.assertEqual([1, 2], res.autoincrement_ids, "autoincrement ok")

    @ensure_version(min=(2, 0))
    async def test__sql_insert_multiple(self):
        res = await self.conn.execute(
            "insert into sql_space (id, name) " "values (1, 'one'), (2, 'two')"
        )
        self.assertEqual(2, res.rowcount, "rowcount ok")

    @ensure_version(min=(2, 0))
    async def test__sql_update(self):
        await self.conn.execute("insert into sql_space values (1, 'one')")

        res = await self.conn.execute("update sql_space set name = 'uno' where id = 1")

        self.assertEqual(1, res.rowcount, "rowcount ok")

    @ensure_version(min=(2, 0))
    async def test__sql_update_multiple(self):
        await self.conn.execute("insert into sql_space values (1, 'one')")
        await self.conn.execute("insert into sql_space values (2, 'two')")

        res = await self.conn.execute("update sql_space set name = 'uno'")

        self.assertEqual(2, res.rowcount, "rowcount ok")

    @ensure_version(min=(2, 0))
    async def test__sql_delete(self):
        await self._compat(self.conn)

        await self.conn.execute("insert into sql_space values (1, 'one')")
        res = await self.conn.execute("delete from sql_space where name = 'one'")
        self.assertEqual(1, res.rowcount, "rowcount ok")

    @ensure_version(min=(2, 0))
    async def test__sql_select(self):
        await self._compat(self.conn)

        await self.conn.execute("insert into sql_space values (1, 'one')")
        await self.conn.execute("insert into sql_space values (2, 'two')")

        res = await self.conn.execute("select * from sql_space")
        self.assertEqual(2, res.rowcount, "rowcount is surely ok")
        self.assertEqual(1, res.body[0][self._compat_field_name("id")])
        self.assertEqual("one", res.body[0][self._compat_field_name("name")])
        self.assertEqual(2, res.body[1][self._compat_field_name("id")])
        self.assertEqual("two", res.body[1][self._compat_field_name("name")])

    @ensure_version(min=(2, 0))
    async def test__sql_delete_multiple(self):
        await self._compat(self.conn)

        await self.conn.execute("insert into sql_space values (1, 'one')")
        await self.conn.execute("insert into sql_space values (2, 'two')")

        res = await self.conn.execute("delete from sql_space")
        self.assertEqual(2, res.rowcount, "rowcount ok")

        res = await self.conn.execute("select * from sql_space")
        self.assertEqual(0, res.rowcount, "rowcount is surely ok")

    @ensure_version(min=(2, 0))
    async def test__metadata(self):
        res = await self.conn.execute("select 1, 2")
        self.assertIsNotNone(res.metadata)
        self.assertIsNotNone(res.metadata.fields)
        self.assertEqual(2, len(res.metadata.fields))

        self.assertEqual("COLUMN_1", res.metadata.fields[0].name)
        self.assertEqual("integer", res.metadata.fields[0].type)
        self.assertEqual("COLUMN_2", res.metadata.fields[1].name)
        self.assertEqual("integer", res.metadata.fields[1].type)

    @ensure_version(min=(2, 0))
    async def test__metadata_names(self):
        await self._compat(self.conn)

        res = await self.conn.execute("select 1 as a, 2 as b")
        self.assertIsNotNone(res.metadata)
        self.assertIsNotNone(res.metadata.fields)
        self.assertEqual(2, len(res.metadata.fields))

        self.assertEqual(self._compat_field_name("a"), res.metadata.fields[0].name)
        self.assertEqual("integer", res.metadata.fields[0].type)
        self.assertEqual(self._compat_field_name("b"), res.metadata.fields[1].name)
        self.assertEqual("integer", res.metadata.fields[1].type)

    @ensure_version(min=(2, 0))
    async def test__metadata_actual_space(self):
        await self._compat(self.conn)

        await self.conn.execute("insert into sql_space values (1, 'one')")
        await self.conn.execute("insert into sql_space values (2, 'two')")

        res = await self.conn.execute("select * from sql_space")
        self.assertEqual(2, res.rowcount, "rowcount is ok")
        self.assertEqual(2, len(res.metadata.fields))
        self.assertEqual(self._compat_field_name("id"), res.metadata.fields[0].name)
        self.assertEqual("integer", res.metadata.fields[0].type)
        self.assertIsNone(res.metadata.fields[0].is_nullable)
        self.assertIsNone(res.metadata.fields[0].is_autoincrement)
        self.assertIsNone(res.metadata.fields[0].collation)

        self.assertEqual(self._compat_field_name("name"), res.metadata.fields[1].name)
        self.assertEqual("string", res.metadata.fields[1].type)
        self.assertIsNone(res.metadata.fields[1].is_nullable)
        self.assertIsNone(res.metadata.fields[1].is_autoincrement)
        self.assertIsNone(res.metadata.fields[1].collation)

    @ensure_version(min=(2, 0))
    async def test__sql_select_full_metadata(self):
        await self._compat(self.conn)

        await self.conn.execute("insert into sql_space values (1, 'one')")
        await self.conn.execute("insert into sql_space values (2, 'two')")

        await self.conn.update(
            "_session_settings", ["sql_full_metadata"], [("=", "value", True)]
        )

        try:
            res = await self.conn.execute("select * from sql_space")
            self.assertEqual(2, len(res.metadata.fields))
            self.assertEqual(self._compat_field_name("id"), res.metadata.fields[0].name)
            self.assertEqual("integer", res.metadata.fields[0].type)
            self.assertEqual(False, res.metadata.fields[0].is_nullable)
            self.assertEqual(None, res.metadata.fields[1].is_autoincrement)
            self.assertIsNone(res.metadata.fields[0].collation)

            self.assertEqual(
                self._compat_field_name("name"), res.metadata.fields[1].name
            )
            self.assertEqual("string", res.metadata.fields[1].type)
            self.assertEqual(True, res.metadata.fields[1].is_nullable)
            self.assertEqual(None, res.metadata.fields[1].is_autoincrement)
            self.assertEqual("unicode", res.metadata.fields[1].collation)
        finally:
            await self.conn.update(
                "_session_settings", ["sql_full_metadata"], [("=", "value", False)]
            )

    async def _compat(self, conn: asynctnt.Connection):
        if conn.version >= (2, 11):
            await conn.execute('SET SESSION "sql_seq_scan" = true;')



import asyncio
from typing import Any, Awaitable, Dict, List, Tuple, Union

from asynctnt.iproto import protocol

MethodRet = Union[Awaitable[protocol.Response], asyncio.Future]
SpaceType = Union[str, int]
IndexType = Union[str, int]
KeyType = Union[List[Any], Tuple]
TupleType = Union[List[Any], Tuple, Dict[str, Any]]
